/* The Java RTP/I library, Version 0.1 alpha.
 *  
 * This library provides the functionality of RTP/I as it is specified in
 * Internet Draft draft-mauve-rtpi-00.txt. 
 *
 * Copyright (C) 2000 Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Martin Mauve
 *
 * e-mail:
 * mauve@informatik.uni-mannheim.de
 * 
 * paper mail:
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi;

/**
 * This interface must be implemented by a class the wants to be contacted
 * by Rtpi when thing happen (e.g. ADUs are received, RTCP/I information arrives,
 * etc).
 */

public interface RtpiRecipient {
    
    /**
     * This method is invoked when the reliability service has detected
     * a lost packet and was not instructed to repair the loss of this
     * packet.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param participantID The ID of the sender of the lost ADU.
     * @param subID The ID of the affected subcomponent.
     * @param type The type (STATE, EVENT, etc.) of the lost ADU.
     * @param sequenceNumber The sequence number of the lost ADU.
     * @param timestamp A timestamp that is greater or equal to the timestamp
     *                  of the lost ADU.
     */

    void rtpiAduLost(Rtpi rtpi, int participantID, long subID, int type, int sequenceNumber, int timestamp);

    /**
     * This method is invoked when the reliability service has detected
     * a lost packet and was instructed to repair the loss but was unable 
     * to do so.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param participantID The ID of the sender of the lost ADU.
     * @param subID The ID of the affected subcomponent.
     * @param type The type (STATE, EVENT, etc.) of the lost ADU.
     * @param sequenceNumber The sequence number of the lost ADU.
     * @param timestamp A timestamp that is greater or equal to the timestamp
     *                  of the lost ADU.
     */

    void rtpiCouldNotRecover(Rtpi rtpi, int participantID, long subID, int type, int sequenceNumber, int timestamp);

    /**
     * This method is called if either the reliability service used for RTP/I or the
     * transport service used for RTCP/I has encountered a fatal error.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     */

    void connectionClosed(Rtpi rtpi);

    /**
     * This method is called to deliver an event that has been received.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param event The event.
     */

    void receiveEvent(Rtpi rtpi, RtpiEvent event);

    /**
     * This method is called to deliver a state that has been received.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param state The state.
     */

    void receiveState(Rtpi rtpi, RtpiState state);

    /**
     * This method is called to deliver a delta state that has been received.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param dletaState The delta event.
     */

    void receiveDeltaState(Rtpi rtpi, RtpiDeltaState deltaState);

    /**
     * This method is called to deliver a state query that has been received.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param stateQuery The state query.
     */

    void receiveStateQuery(Rtpi rtpi, RtpiStateQuery stateQuery);


    /**
     * This method is called when the information for a participant has changed.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param participant The new participant information.
     */

    void changeSourceInfo(Rtpi rtpi, RtpiSourceInfo participant);

    /**
     * This method is called when a participant has timed out or sent a BYE.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param participant The participant information about the removed participant.
     */

    void removeSource(Rtpi rtpi, RtpiSourceInfo participant);

    /**
     * This method is called when a new subcomponent has been discovered by RTCP/I. If the subcomponent
     * has been added by the local application to the list of those subcomponents that it is
     * interested in (by using Rtpi.addSubcomponent) and has then removed the subcomponent from
     * that list (using Rtpi.removeSubcomponent) then this method MAY be called if some other application
     * still announces the subcomponent. If this is important to the application, then it should remember
     * the subcomponents it has seen so far and ignore this message if the subcomponent has already been
     * seen.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param subcomponentID The ID of the subcomponent.
     * @param active True if the subcomponent is active.
     * @param name The name of the subcomponent.
     */

    void addSubcomponent(Rtpi rtpi, long subcomponentID, boolean active, byte[] name);

    /**
     * This method is called when a subcomponent has timed out. This can only happen
     * when the local application has not added the subcomponent via Rtpi.addSubcomponent.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param subcomponentID The ID of the subcomponent.
     */

    void removeSubcomponent(Rtpi rtpi, long subcomponentID);

    /**
     * This method is called when a subcomponent has become active. This is only called when
     * the subcomponent has not been declared as active by the local application. 
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param subcomponentID The ID of the subcomponent.
     */

    void activateSubcomponent(Rtpi rtpi, long subcomponentID);

    /**
     * This method is called when a subcomponent has become passive.
     *
     * @param rtpi The Rtpi instance that generated this message. This is usefull if multiple
     *             Rtpi instances are used by a single application.
     * @param subcomponentID The ID of the subcomponent.
     */

    void deactivateSubcomponent(Rtpi rtpi, long subcomponentID);

}
