/* The Java RTP/I library, Version 0.1 alpha.
 *  
 * This library provides the functionality of RTP/I as it is specified in
 * Internet Draft draft-mauve-rtpi-00.txt. 
 *
 * Copyright (C) 2000 Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Martin Mauve
 *
 * e-mail:
 * mauve@informatik.uni-mannheim.de
 * 
 * paper mail:
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.packets;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;

/**
 * This implements the RTCP/I bye packet.
 */

public class RtcpiByePacket extends RtcpiPacket {
    
    /**
     * The payload type for bye packets.
     */

    public static final int BYE   = 71;
    byte[] reason=null;

    /**
     * This creates a new bye packet that is to be transmitted over 
     * the network.
     *
     * @param pid The id of the participant transmitting the packet.
     */

    public RtcpiByePacket(int pid) {
	super(pid);
	payloadType=BYE;
	length=0;
    }

    /**
     * This creates a new bye packet that is to be transmitted over 
     * the network.
     *
     * @param pid The id of the participant transmitting the packet.
     * @param r The reason for leaving the session.
     */

    public RtcpiByePacket(int pid, byte[] r) {
	super(pid);
	payloadType=BYE;
	reason=r;
	length=reason.length+1;
	length=length+(4-length%4)%4;
	length/=4;
    }

    /**
     * This creates an RTCP/I bye packet from a transport packet
     * that has been received.
     *
     * @param d The transport packet.
     * @param s The start of the RTCP/I packet in the transport packet 
     *          (position of the first byte of the RTCP/I header.
     */

    public RtcpiByePacket(byte[] d, int s) {
	super(d, s);
    }

    /**
     * This flushes an outgoing packet.
     */

    public void flush() throws RtpiFlushException {
	ByteArrayOutputStream bos = new ByteArrayOutputStream();
	DataOutputStream dos = new DataOutputStream(bos);
	
	flushHeader(dos);

	if (reason!=null) {
	    if (reason.length>255) {
		throw new RtpiFlushException("RtcpiByePacket: Reason too long");
	    }
	    try {
		dos.writeByte((byte) reason.length);
		dos.write(reason);
	    } catch (Exception ex) {
		throw new RtpiFlushException("RtcpiByePacket: Couldn't write reason");
	    }
	}
	
	try {
	    dos.flush();
	    bos.flush();
	} catch (Exception ex) {
	    throw new RtpiFlushException("RtcpiByePacket: Couldn't flush output streams");
	}
	
	packetData = bos.toByteArray();
	if (packetData.length!=(length+2)*4) {
	    throw new RtpiFlushException("RtcpiByePacket: Illegal packet length after flushing packet");
	}

	try {
	    dos.close();
	    bos.close();
	} catch (Exception ex) {
	    throw new RtpiFlushException("RtcpiByePacket: Couldn't close output streams");
	}
    }
    
    /**
     * This parses an incoming packet.
     */

    public void parse() throws RtpiParseException {
	ByteArrayInputStream bis = new ByteArrayInputStream(packetData, packetStart, packetData.length-packetStart);
	DataInputStream dis = new DataInputStream(bis);
	
	parseHeader(dis);

	if (payloadType!=BYE) {
	    throw new RtpiParseException("RtcpiByePacket:wrong payload type, was "+payloadType);
	}

	if (length!=0) {
	    try {
		int reasonLength=dis.readUnsignedByte();
		reason=new byte[reasonLength];
		dis.readFully(reason);
	    } catch (Exception ex) {
		throw new RtpiParseException("RtcpiByePacket: could not read reason");
	    }
	}

	try {
	    dis.close();
	    bis.close();
	} catch (Exception exe) {
	    throw new RtpiParseException("RtcpiByePacket: Could not close input stream for packet");
	}
    }

    /**
     * This returns the reason for transmitting the bye packet.
     *
     * @return The reason for transmitting the bye packet.
     */

    public byte[] getReason() {
	return reason;
    }

    /**
     * Convert this object to a string.
     *
     * @return This object as a string.
     */
    
    public String toString() {
	String returnValue;
	returnValue ="RtcpiByePacket: {\n";
	returnValue+=toStringDetails();
	returnValue+="}\n";
	return returnValue;
    }


    String toStringDetails() {
	String returnValue;
	returnValue=super.toStringDetails();
	if (reason!=null) {
	    returnValue+="  Reason for leaving: "+new String(reason)+"\n";
	}
	return returnValue;
    }
}

