/* The Java RTP/I library, Version 0.1 alpha.
 *  
 * This library provides the functionality of RTP/I as it is specified in
 * Internet Draft draft-mauve-rtpi-00.txt. 
 *
 * Copyright (C) 2000 Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Martin Mauve
 *
 * e-mail:
 * mauve@informatik.uni-mannheim.de
 * 
 * paper mail:
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.packets;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * This is the base class for RTCP/I packets. 
 *
 * @author Martin Mauve
 */

public abstract class RtcpiPacket {

    /**
     * The size of an RTCP/I header.
     */

    public static final int HEADER_SIZE        = 8;

    /**
     * The RTCP/I version number.
     */

    public final static int VERSION = 0;


    // Bitmasks for parsing incoming RTCP/I packets:

    /**
     * Bitmask for the version number.
     */
 
    private final static int VERSION_BITMASK                    = 0xC0;

    /**
     * Bitmask for the reserved bit.
     */
 
    private final static int RESERVED_BITMASK                    = 0x20;

    /**
     * Bitmask for the count field.
     */
 
    private final static int COUNT_BITMASK     = 0x1F;


    // Bitmasks for removing the sign

    /**
     * Bitmask for transforming signed to unsigned value.
     */

    private final static int BYTE_REMOVE_SIGN_BITMASK   = 0x7f;

    /**
     * Bitmask for transforming signed to unsigned value.
     */

    private final static int BYTE_SIGN_BITMASK          = 0x80;

    /**
     * Bitmask for transforming signed int to unsigned value.
     */

    private final static int INT_REMOVE_SIGN_BITMASK     = 0x7fffffff;

    /**
     * Bitmask for transforming signed int to unsigned value.
     */

    private final static int INT_SIGN_BITMASK            = 0x80000000;


    // RCTP/I header fields

    /**
     * The version.
     */

    protected int version=VERSION;

    /**
     * The reserved bit.
     */

    protected int reserved=0;

    /**
     * The count field.
     */

    protected int count=0;

    /**
     * The RTCP/I payload type (SUBREP / APP / BYE / SDES).
     */

    protected int payloadType=0;

    /**
     * The length of this RTCP/I packet in bytes.
     */

    protected int length=0;

    /**
     * The participant ID
     */

    protected int participantID=0;

    // RTCP/I Transport Packet Information:

    /**
     * RTCP packets are stacked in one compound packet.
     * We therefore need to remember where it does start
     * within the transport packet!
     */

    protected int packetStart=0;

    /**
     * The whole compound packet (transport packet) which
     * contains this RTCP packet.
     */

    protected byte[] packetData=null;

    /**
     * This static method extracts the RTCP/I payload type of an
     * RTCP/I packet.
     *
     * @param packet The packet which contains the RTCP/ packet.
     * @param start The start (position of the first byte of the header)
     *              of the RTCP/I packet in the transport packet.
     * @return The payload type.
     */

    public static int inspectPayloadType (byte[] packet, int start) 
	throws RtpiParseException {
	
	if (packet.length < start+2) {
	    throw new RtpiParseException("RtcpiPacket: Could not access payloadType byte for determining the type of the RTCP/I packet");
	}

	byte pt=packet[start+1];	

	int leadingBit = (pt & BYTE_SIGN_BITMASK) >>> 7;
	int retVal=pt&BYTE_REMOVE_SIGN_BITMASK;
	retVal=retVal|(leadingBit<<7);
	return retVal;
    }


    /**
     * This static method extracts the length of the
     * packet.
     *
     * @param packet The packet which contains the RTCP/ packet.
     * @param start The start (position of the first byte of the header)
     *              of the RTCP/I packet in the transport packet.
     * @return The length of the packet including the RTCP/I header in bytes.
     */

    public static int inspectLength (byte[] packet, int start) 	throws RtpiParseException {

	if (packet.length < start+4) {
	    throw new RtpiParseException("RtcpiPacket: Could not access length for determining the size of the RTCP/I packet");
	}

	byte b=packet[start+2];

	int leadingBit = (b & BYTE_SIGN_BITMASK) >>> 7;
	int retVal=b&BYTE_REMOVE_SIGN_BITMASK;
	retVal=(retVal|(leadingBit<<7))<<8;

	b = packet[start+3];

	leadingBit = (b & BYTE_SIGN_BITMASK) >>> 7;
	int help = b&BYTE_REMOVE_SIGN_BITMASK;
	help=(help|(leadingBit<<7));

	retVal = retVal|help;
 
	return retVal*4+HEADER_SIZE;
    }

    /**
     * Constructor for packets coming from the network
     */

    protected RtcpiPacket(byte[] packet, int start) {
	packetData=packet;
	packetStart=start;
    }

    /**
     * Constructer for packets that are to be sent to the network
     */

    protected RtcpiPacket(int pid) {
	participantID=pid;
    }

    /**
     * This returns the encoded packet.
     *
     * @return The encoded packet.
     */

    public byte[] getPacket() {
	return packetData;
    }

    /**
     * This returns the Partcicipant ID of this packet's source.
     *
     * @return The Participant ID.
     */

    public int getParticipantID() {
	return participantID;
    }

    /**
     * This returns the length of this packet.
     *
     * @return The length.
     */

    public int getLength() {
	return length;
    }

    /**
     * Get the payload type of the RTCP packet.
     *
     * @return The payload type.
     */

    public int getPayloadType() {
	return payloadType;
    }

    /**
     * This flushes the packet. To be overidden in derived
     * classes. 
     *
     * @return The flushed packet data.
     */

    public abstract void flush() throws RtpiFlushException;

    /**
     * This flushes the header part of an RTCP/I packet. 
     *
     * @param dos The output stream to which the header information should be flushed.
     * @return The flushed header.
     */
    
    protected void flushHeader(DataOutputStream dos) throws RtpiFlushException {
	int word;

	word=version<<6;
	word=word|(reserved<<5);
	word=word|count;

	try {
	    dos.writeByte((byte)word);
	} catch (Exception ex) {
	    throw new RtpiFlushException("RtpiPacket: Couldn't write first byte to RTCP/I header!");
	}

	try {
	    dos.writeByte((byte) payloadType);
	} catch (Exception ex) {
	    throw new RtpiFlushException("RtpiPacket: Couldn't write payload type to RTCP/I header");
	}

	try {
	    dos.writeShort((short) length);
	} catch (Exception ex) {
	    throw new RtpiFlushException("RtpiPacket: Couldn't write length to RTCP/I header");
	}

	try {
	    dos.writeInt(participantID);
	} catch (Exception ex) {
	    throw new RtpiFlushException("RtpiPacket: Couldn't write participant ID to RTCP/I header");
	}

    }
	
    
    /**
     * This parses the packet. To be overidden in derived
     * classes. 
     */

    public abstract void parse() throws RtpiParseException;

    /**
     * This parses the header part of an RTCP packet. 
     *
     * @param dis The input stream from which the information can be read.
     */
    
    void parseHeader(DataInputStream dis) throws RtpiParseException {
	int word=0;

	try {
	    word = dis.readUnsignedByte();
	} catch (Exception exe) {
	    throw new RtpiParseException("RtpiPacket: Could not read first byte from RTCP/I header");
	}
	
	version=(word & VERSION_BITMASK) >>> 6;
	reserved=(word & RESERVED_BITMASK) >>> 5;
	count=word & COUNT_BITMASK;

	if (version != VERSION) {
	    throw new RtpiParseException("RtpiPacket: Incorrect Version number in RTCP/I header");
	}

	try {
	    payloadType = dis.readUnsignedByte();
	} catch (Exception exe) {
	    throw new RtpiParseException("RtpiPacket: Could not read payload type from RTCP/I header");
	}

	try {
	    length = dis.readUnsignedShort();
	} catch (Exception exe) {
	    throw new RtpiParseException("RtpiPacket: Could not read packet length from RTCP/I header");
	}
	
	try {
	    participantID = dis.readInt();
	} catch (Exception exe) {
	    throw new RtpiParseException("RtpiPacket: Could not read participant ID from RTCP/I header");
	}

    }

    /**
     * Convert this object to a string.
     *
     * @return This object as a string.
     */
    
    public String toString() {
	String returnValue;
	returnValue ="RTCPIPacket: {\n";
	returnValue+=toStringDetails();
	returnValue+="}\n";
	return returnValue;
    }


    String toStringDetails() {
	String returnValue;
	returnValue ="  version: "+version+"\n";
	returnValue+="  reserved: "+reserved+"\n";
	returnValue+="  count: "+count+"\n";
	returnValue+="  payloadType: "+payloadType+"\n";
	returnValue+="  length: "+length+"\n";
	returnValue+="  Participant ID: "+participantID+"\n";	
	returnValue+="  packetStart: "+packetStart+"\n";
	return returnValue;
    }

}

