/* The Java RTP/I library, Version 0.1 alpha.
 *  
 * This library provides the functionality of RTP/I as it is specified in
 * Internet Draft draft-mauve-rtpi-00.txt. 
 *
 * Copyright (C) 2000 Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Martin Mauve
 *
 * e-mail:
 * mauve@informatik.uni-mannheim.de
 * 
 * paper mail:
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.reliability;

import rtpi.IllegalValueException;
import java.util.LinkedList;


/**
 * This describes the interface to a reliability service that
 * is used together with RTP/I.
 */

public interface Reliable {

    /**
     * The NONE quality of service is used when the reliability
     * service should take no actions to notice or repair lost
     * ADUs of the given subcomponent and type pair.
     */

    public static final int NONE=0;

    /**
     * The DETECT quality of service is used when the reliability
     * service should detect lost ADUs but should not try to repair 
     * lost ADUs.
     */

    public static final int DETECT=1;

    /**
     * The RELIABLE quality of service is used when the reliability
     * service should detect and repair lost ADUs of a give subcomponent
     * and type pair. Notice that a reliability service may not be able 
     * to recover an ADU and, in this case, will notify the recipient
     * of the lost ADU that it was unable to recover the ADU.
     */

    public static final int RELIABLE=2;

    /**
     * As with RELIABLE the ORDERED quality of service is used to request
     * the reliable reception of ADUs of a given subcomponent and type
     * pair. However, the ADUs are deliverd in the same order in which
     * they have been transmitted. This is source ordering. It does make
     * no assertments to retain the order of ADUs that have been sent
     * by more than one source.
     */

    public static final int ORDERED=3;

    /**
     * This method is used to transmit an ADU. The ADU may consist of
     * multiple RTP/I packets if the ADU is too large to fit into a
     * single (network) packet. The RTP/I packets MUST form a single
     * valid RTP/I ADU.
     * As an option the ADU may be transmitted with a certain amount of 
     * redundancy that is to be transmitted in a certain time interval 
     * to compensate for packet loss. It is important to notice that a 
     * given implementation may not support redundancy. 
     *
     * @param packet A linked list of the RTP/I packets that belong to a single ADU.
     * @param redundancy The amount of redundancy that should be added to this ADU.
     * @param transmissionInterval The interval during which redundancy information
     *                             may be transmitted.
     */

    void transmitRtpiAdu(LinkedList packets, 
			 float redundancy, 
			 int transmissionInterval)
	throws QosNotSupportedException, IllegalValueException;
    

    /**
     * This method is used to set the QoS for a given subcomponent and
     * RTP/I ADU type pair. It is important to notice that a given implementation
     * of the Reliable interface may only support certain QoS settings.
     *
     * @param subcomponentID The ID of the subcomponent.
     * @param type The RTP/I ADU type.
     * @param qos The desired QoS.
     */

    void setInterest(long subcomponentID,
		     int type,
		     int qos)
	throws QosNotSupportedException;
    
    /**
     * This method is used to set a default QoS for all subcomponent and
     * RTP/I ADU type pairs to which no other QoS has been specified. 
     * It is important to notice that a given implementation
     * of the Reliable interface may only support certain QoS settings.
     *
     * @param subcomponentID The ID of the subcomponent.
     * @param type The RTP/I ADU type.
     * @param qos The desired QoS.
     */

    void setDefaultInterest(int type, int qos)
	throws QosNotSupportedException;

    /**
     * This joins the group used by the reliability service
     * to communicate with peer instances (e.g. a multicast
     * group). This MUST be called before any ADUs can be
     * transmitted or received.
     */
 
    void joinGroup();
    
    /** 
     * This leaves the group used by the reliability service
     * to communicate with peer instances (e.g. a multicast
     * group). After this method has been called no ADUs may
     * be transmitted and no further ADUs can be received.
     * However, already received ADUs may still be delivered.
     */
 
    void leaveGroup();

    /**
     * This registers the recipient of the ADUs and
     * loss notifications.
     * 
     * @param rec The reipient.
     */

    void registerRecipient(ReliableRecipient rec);

    /**
     * This terminates the reliability service.
     */

    void quit();


    /**
     * This returns the size of the payload that can
     * be transported in a single packet. This payload
     * size INCLUDES all RTP/I and reliability headers.
     *
     * @return The transport payload size.
     */

    int getTransportPayloadSize();
    
    /**
     * This returns the size of the combined RTP/I and
     * reliability headers. 
     *
     * @return The header size.
     */

    int getCombinedHeaderSize();
}
