/* The Java RTP/I latejoin service, Version 0.1 alpha.
 *  
 * This library provides the functionality of a generic RTP/I latejoin service.
 *
 * Copyright (C) 2000 Juergen Vogel, Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Juergen Vogel
 * Martin Mauve
 *
 * e-mail:
 * {vogel,mauve}@informatik.uni-mannheim.de
 * 
 * paper mail:
 *
 * Juergen Vogel
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */


package rtpi.services.latejoin;

import rtpi.transport.Transport;
import rtpi.transport.TransportRecipient;
import rtpi.transport.TransportPacket;
import rtpi.transport.ipmc.IPMCTransport;

import rtpi.RtpiSourceInfo;

import java.net.*;
import java.io.*;


class ControlChannelManager implements TransportRecipient {
    private static final boolean debug = false;

    private Transport controlChannel;
    private RtpiSourceInfo localParticipant;

    private LateJoin lj;

    ControlChannelManager(LateJoin lateJoin, Transport cChannel, RtpiSourceInfo participant) {
	lj = lateJoin;
	localParticipant = participant;

	controlChannel = cChannel;
	controlChannel.registerTransportRecipient(this);
	try {
	    controlChannel.joinGroup();
	} catch (IOException ex) {
	    System.err.println("ControlChannelManager::ControlChannelManager " +
			       " - Exception while setting up control channel " + ex);
	}
    }

    public void receiveTransportPacket(TransportPacket packet) { // control packet
	lj.messages().put(new LateJoinMessage(LateJoinMessage.RECEIVE_CONTROL_PACKET,
					      packet));

    }

    void execReceiveTransportPacket(TransportPacket packet) { // control packet
	ControlPacket controlPacket = new ControlPacket(packet);

	if (controlPacket.getGenericServiceType() != ControlPacket.LATE_JOIN) {
	    System.err.println("ControlChannelManager::receiveControlPacket - wrong generic service type");
	    return;
	}


	// STATE QUERY
	if (controlPacket.getType() == ControlPacket.STATE_QUERY) {
	    lj.getSubcomponentManager().handleStateQuery(controlPacket.getSubcomponentID(),
							 controlPacket.getParticipantID(),
							 controlPacket.getTries());

	    if (debug)
		System.out.println("ControlChannelManager::receiveControlPacket - " +
				   "received State Query for Subcomponent with ID " +
				   controlPacket.getSubcomponentID() +
				   " , try: " + controlPacket.getTries());
	    return;
	}

	// SERVER ACKNOWLEDGEMENT
	if (controlPacket.getType() == ControlPacket.ACK_SERVER) {

	    receiveServerSelectionAck(controlPacket.getSubcomponentID());

	    if (debug)
		System.out.println("ControlChannelManager::receiveControlPacket - " +
				   "received Server Acknowledgement for Subcomponent with ID " +
				   controlPacket.getSubcomponentID());
	    return;
	}

	System.err.println("ControlChannelManager::receiveControlPacket - wrong type");
    }

    void requestSubcomponentState(long id, int t) {
	ControlPacket packet = new ControlPacket();
	packet.setGenericServiceType(ControlPacket.LATE_JOIN);
	packet.setType(ControlPacket.STATE_QUERY);
	packet.setParticipantID(localParticipant.getParticipantID());
	packet.setSubcomponentID(id);
	packet.setTries(t);

	transmitControlPacket(packet);

 	if (debug)
	    System.out.println("ControlChannelManager::transmitStateQuery for Subcomponent with id " +
			       id + " , try: " + t);
    }

    void acknowledgeServerSelection(long id) {
	ControlPacket packet = new ControlPacket();
	packet.setGenericServiceType(ControlPacket.LATE_JOIN);
	packet.setType(ControlPacket.ACK_SERVER);
	packet.setParticipantID(localParticipant.getParticipantID());
	packet.setSubcomponentID(id);

	transmitControlPacket(packet);

 	if (debug)
	    System.out.println("ControlChannelManager::acknowledgeServerSelection " +
			       "for Subcomponent with id " + id);
    }

    private void transmitControlPacket(ControlPacket packet) {
	controlChannel.sendTransportPacket(packet.getEncodedPacket());
    }

    private void receiveServerSelectionAck(long id) {
	SubcomponentManager sm = lj.getSubcomponentManager();
	Subcomponent ljSub = sm.getSubcomponent(id);
	TimerManager tm = lj.getTimerManager();

	if (ljSub == null) {
	    if (debug)
		System.out.println("ControlChannelManager::receiveServerSelectionAck - " +
				   "Subcomponent with ID " + id + " not discovered yet");
	    return;
	}

	ljSub.setLastTimeSeenNow();

	if (tm.stateQueryTimerRunning(id)) {
	    tm.setStateQueryAcknowledged(id);

	    if (debug)
		System.out.println("ControlChannelManager::receiveServerSelectionAck - " +
				   "requested Subcomponent with ID " + id + " will come soon");
	    return;
	}

	if (tm.stateReplyTimerRunning(id)) {
	    tm.deleteStateReplyTimer(id);

	    if (debug)
		System.out.println("ControlChannelManager::receiveServerSelection - " +
				   "delete StateReply-Timer for Subcomponent with ID " + id);
	    return;
	}

	// if preparing to send: to avoid duplicates, abort if our ip-address is higher

    }

    void quit() {
	controlChannel.leaveGroup();
    }

    public void connectionClosed() {}

}

