/* The Java RTP/I latejoin service, Version 0.1 alpha.
 *  
 * This library provides the functionality of a generic RTP/I latejoin service.
 *
 * Copyright (C) 2000 Juergen Vogel, Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Juergen Vogel
 * Martin Mauve
 *
 * e-mail:
 * {vogel,mauve}@informatik.uni-mannheim.de
 * 
 * paper mail:
 *
 * Juergen Vogel
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */


package rtpi.services.latejoin;

import java.util.Hashtable;
import java.util.Enumeration;

class DelayValue {
    long delay;

    DelayValue(long d) {
	delay = d;
    }
}


class DelayManager {
    private static boolean debug = false;
    private static final double ALPHA = 0.6;
    private static final long DEFAULT = 100;
    private static final long MAX_DELAY = 300;

    private Hashtable delayTable;


    DelayManager() {
	delayTable = new Hashtable();
    }

    void addParticipant(int lssrc) {
	Integer ssrc = new Integer(lssrc);
	if (delayTable.get(ssrc) != null) return; // already exists

	delayTable.put(ssrc, new DelayValue(DEFAULT));
    }

    boolean participantExists(int pid) {
	return delayTable.containsKey(new Integer(pid));
    }

    void removeParticipant(int lssrc) {
	Integer ssrc = new Integer(lssrc);
	if (delayTable.get(ssrc) == null) return; // does not exist

	delayTable.remove(ssrc);
    }

    void addDelayValue(int lssrc, long timestamp) {
	Integer ssrc = new Integer(lssrc);
	long currentDelay = (System.currentTimeMillis()&0x00000000FFFFFFFFl) - timestamp;
	
	if (currentDelay < 0) { //This prevents that participants with unsynchronized clocks destroy the time calculation
	    currentDelay*=-1;
	}
	if (currentDelay>MAX_DELAY) { // This prevents that participants with unsynchronized clocks destroy the time claculation
	    currentDelay=MAX_DELAY;
	}

	DelayValue delay = (DelayValue)delayTable.get(ssrc);
	if (delay == null) {  // does not exist
	    addParticipant(lssrc);
	    delay = (DelayValue)delayTable.get(ssrc);
	}

	if (delay.delay == DEFAULT)
	    delay.delay = currentDelay;
	else
	    delay.delay = Math.round(ALPHA*(double)currentDelay + (1 - ALPHA)*delay.delay);

	if (debug)
	    System.out.println("DelayManager::addDelayValue - Participant (SSRC): " + ssrc +
			       " delay: " + delay.delay);
    }

    long getDelay(int ssrc) {
	DelayValue delay = (DelayValue)delayTable.get(new Integer(ssrc));

	if (delay == null) return DEFAULT;
	else return delay.delay;
    }

    long getAverageDelay() {
	if (delayTable.size() == 0) // no entries yet
	    return DEFAULT;

	double average = 0.0;

	for (Enumeration e = delayTable.elements(); e.hasMoreElements(); )
	    average += ((DelayValue)e.nextElement()).delay;

	average /= delayTable.size();

	return Math.round(average);
    }

    long getMaxDelay() {
	if (delayTable.size() == 0) // no entries yet
	    return DEFAULT;

	Enumeration e = delayTable.elements();

	long max = ((DelayValue)e.nextElement()).delay;
	long value;

	while (e.hasMoreElements()) {
	    value = ((DelayValue)e.nextElement()).delay;
	    if (max < value)
		max = value;
	}

	return max;
    }
}
