/* The Java RTP/I latejoin service, Version 0.1 alpha.
 *  
 * This library provides the functionality of a generic RTP/I latejoin service.
 *
 * Copyright (C) 2000 Juergen Vogel, Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Juergen Vogel
 * Martin Mauve
 *
 * e-mail:
 * {vogel,mauve}@informatik.uni-mannheim.de
 * 
 * paper mail:
 *
 * Juergen Vogel
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.services.latejoin;

import rtpi.Rtpi;
import rtpi.RtpiRecipient;
import rtpi.RtpiEvent;
import rtpi.RtpiState;
import rtpi.RtpiDeltaState;
import rtpi.RtpiStateQuery;
import rtpi.RtpiSourceInfo;

import rtpi.transport.Transport;
import rtpi.transport.TransportPacket;

import rtpi.util.SyncQueue;

/**
 * The class LateJoin is the main API interface for using the generic late join service. 
 *
 * <P>The main idea of this generic late join service is to use a separate RTP/I session
 * for requesting and transmitting late join information. This session is called the late join RTP/I session.
 * The late join RTP/I session must be provided by the application and should operate on a <b>network</b> 
 * address that is distinct from the base RTP/I session which is used for the regular transmission of media
 * ADUs (e.g. events and states). The application is responsible for creating and terminating the
 * late join RTP/I session as well as setting QoS parameters. The late join service will join and 
 * leave the session as appropriate.
 *
 * <P>The late join service needs to see the messages that arrive from the base RTP/I session. Therefore
 * it will register as recipient to the base RTP/I session. The application in turn is registered as recipient
 * of the late join service. The application must hand outgoing ADUs directly to the base RTP/I session.
 * Incoming ADUs and information (such as loss notifications and RTCP/I information) is forwarded by the late
 * join service to the application that registered as recipient to the late join service.
 *
 * <P>This late join service basically works as follows: whenever a new subcomponent is discovered by means
 * of an RTCP/I subrep packet the application will be queried what to do. The application then sets a certain
 * late join policy for that subcomponent. According to this policy the late join service will try to query
 * the state of the subcomponent at an appropriate time in a scalable way. For example if the late join policy
 * is chosen to be EVENT, then the late join service will monitor incoming ADUs of the base RTP/I session.
 * As soon as an event for the subcomponent is discovered the late join service will try to query the state of the
 * subcomponent using a random timer based query and response suppression sceme similar to that of SRM.
 * It is important for the scalability of the late join service, that the clocks of the participants are synchronized.
 * The usage of NTP on all computers participating in a session is highly recommended!
 *
 * <P>The late join service leaves the late join RTP/I session when all subcomponents have been dealt with according
 * to the application defined policies and when it has not helped with a late join of another participant for a
 * certain period of time. The late join service rejoins the late join RTP/I session when it is required to help
 * another late comer with its late join. In order to be able to do so a late comer will query the state of a
 * subcomponent for a number of times using the late join RTP/I session. If it has not received any reply it is likely
 * that no participant that is member of the late join RTP/I session holds the state of the subcomponent. Therefore the
 * late join service will use a specific Transport session to request that someone should join the late join RTP/I session 
 * who is able to reply to the state query. Every late join service that holds the state of the requested subcomponent and
 * that has received this request performs a drawing of a random number. If the result is below a certain threshold it
 * will join the late join RTP/I session and transmit the requested state. In this way a join implosion is prevented.
 *
 * <P>The late join service is highly customizable to support the different needs of distributed interactive media.
 * Most importantly an application may choose to use the algorithm described above for determining if it should join
 * the RTP/I late join session when a late comer needs the state of a certain sub-component. As an alternative the application
 * may decide for which sub-components it is responsible. For example if a floor control mechanism is used, the floor
 * holder could always be responsible for transmitting the state of a sub-component. To support this the late join service
 * allows the application to decide for which sub-components it is responsible. When a late join service receives a message
 * that the state of this subcomponent is required to perform a late join, then it will automatically join the late join
 * RTP/I session and transmit the state of the sub-component.
 *
 * <P>Furthermore the application may opt to set a default late join policy for all sub-components. In this case the 
 * application is never asked to specify a late join policy. Instead the late join service automatically uses the 
 * default policy.
 * 
 * <P> Please note that it is not the task of the late join service to guarantee consistency. It is expected that the
 * application has some way to monitor and ensure consistency. The late join serves merely provides an initial state
 * for the subcomponents in a scalable and policy driven way.
 *
 * <P>This late join service works in its own thread. The communication with other threads (such as the application) 
 * is done via message passing, so as to avoid dead locks and race conditions. Therefore the methods of the late
 * join service do not have return values. Also, the application should not block any method calls from the late join
 * service. It should either use message passing or some other way to prevent dead locks and race conditions.
 *
 * <P> For publications about RTP/I and related issues, please refer to:
 * http://www.informatik.uni-mannheim.de/informatik/pi4/projects/RTPI/index.html
 * 
 * @author Juergen Vogel
 * @author Martin Mauve
 */


public final class LateJoin extends Thread implements RtpiRecipient {

    // Costants for the controling the behaviour of the late join service:

    /**
     * This signals that no late join policy has been specified for the subcomponent, yet.
     */

    static final int NO_POLICY_YET         = 0;

    /**
     * This policy is used by the application when it does not need the state of the subcomponent.
     */
 
    public static final int NOT_INTERESTED = 1;

    /**
     * This policy instructs the late join service to get the state of the subcomponent as fast as 
     * possible.
     */

    public static final int AT_ONCE        = 2;

    /**
     * With this policy the late join service waits for the first event or delta state it sees for
     * the subcomponent before it tries to recover the state of the subcomponent.
     */

    public static final int EVENT          = 3;


    /**
     * With this policy the late join service waits untill the network load is below a certain threshhold
     * before it requests the state of the subcomponent. This policy has not been implemented, yet.
     */

    static final int NET_CAPACITY   = 4; // not operational yet!

    /**
     * This join policy indicates exclusive server responsibiltiy for a particular
     * subcomponent implicating both replies without reply-implosion-avoidance and instant joining
     * of the late join multicast group. This mode is used for example with floor control.
     */

    public static final int APPLICATION = 0;

    /**
     * This join policy selects servers automatically by computed probabilities. This mode is
     * used when multiple applications can serve the state of subcomponents. The late join service
     * will automatically select an application to reply to late join induced state queries.
     */
 
    public static final int AUTOMATIC   = 1;


    private final boolean debug = false;

    // These are the default values for the optional parameters
    private static final int INITIAL_REQUESTS_DEFAULT = 5;
    private static final int JOIN_REQUESTS_DEFAULT = 10;
    private static final float NO_JOIN_PROBABILITY_DEFAULT = 0.7f;

    // queue to receive messages from other threads
    private SyncQueue messages = new SyncQueue();

    // appplication multicast channel of data group
    private LateJoinRecipient recipient;

    // control multicast channel of data group
    private ControlChannelManager cChannelManager;

    // late join multicast channel
    private LateJoinChannelManager ljChannelManager;

    // subcomponents
    private SubcomponentManager subcomponentManager;

    // timer
    private TimerManager timerManager;

    // client state
    private ClientStateManager clientStateManager;

    // server state
    private MembershipManager membershipManager;

    // delay
    private DelayManager delayManager;

    // base RTP/I session
    Rtpi rtpi;

    // late join RTP/I session
    Rtpi ljRtpi;

    // number of requests we transmit on the late join channel
    int initialRequests;

    // number of requests for joining the late join channel before giving up.
    int joinRequests;

    // probability for finding noone to join the latejoin group, provided that every participant could
    // serve as late join server.
    float joinProbability;

    /**
     * This creates a LateJoin instance in the most flexible way that is possible.
     *
     * @param ljRecipient The recipient of information from this instance. Commonly this is the application.
     * @param r The base RTP/I session. The late join service will not transmit anything over this session. However
     *          it needs to monitor incoming RTP/I ADUs and the RTCP/I information.
     * @param ljR The late join RTP/I session. The late join service uses this session to transmit state queries and
     *            states. This RTP/I session is created and destroyed by the application. The application is also responsible
     *            for the QoS settings of the underlying reliability service. The late join service joins and leaves this session
     *            as it is appropriate.
     * @param ljParticipant This information is used for the RTCP/I SDES reporting in the late join session. It should be minimal.
     *                      That is only the CNAME should be set.
     * @param payloadType This is the payload type use by the application.
     * @param genericServices This is the generic services channel. In the current version this is a Transport session which is 
     *                        exclusively used by the late join service to signal that someone should join the late join RTP/I session
     *                        and provide the state of a certain subcomponent. The application providing this Transport object is responsible
     *                        for creating and cleaning up this Transport instance.
     * @param joinPolicy This may be AUTOMATIC if the lateJoin group should be joined automatically using the random number
     *                   mechanisms described above. Or it can be APPLICATION if the application decides for which subcomponents
     *                   this late join should be responsible. The latter should only be chosen if the responsibility for a subcomponent
     *                   rests with exactly one participant in the session.
     * @param minMembershipTime This is the minimal time that a late join service will stay in the late join RTP/I session once it joins.
     *                          At the startup of the late join service it will automatically join the late join RTP/I session. Once it has
     *                          left the late join RTP/I session it may rejoin in order to help a late comer. 
     * @param maxMembershipTime This is the maximal time tat a late join service will stay in the late join RTP/I session once it joins.
     *                          The late join service will choose a value between the minMembershipTime and the maxMembership time for
     *                          leaving the late join session if it has finished its late join for all discovered subcomponents and if it
     *                          has not transmitted any subcomponent state to help another latecomer.
     * @param maxStateTransmissionTime This is the maximum time required by the application to extract and succesfully transmit the state
     *                                 of a subcomponent. Obviously this has to bee an estimate, since the successfull transmission may require
     *                                 retransmissions of lost packets. If this value is too low than the late join service will repeatedly
     *                                 ask for the state of a subcomponent while a member in the late join RTP/I session is already extracting
     *                                 and transmitting the state. If it is set too large, then it takes longer to discover that a state transmission
     *                                 has got lost, or the sender of the state has crashed before it could finish transmitting the state.
     *                                 Therefore this value should be set to a loarge amount if reliable communication is used for the transmission
     *                                 of states over the late join RTP/I session. 
     * @param defaultPolicy One of NO_POLICY_YET, NOT_INTERESTED, AT_ONCE or EVENT. NO_POLICY_YET signals that there is
     *                      no default policy. The meaning of the other policies is explained in the description of the constants. If the defaultPolicy
     *                      is different from NO_POLICY_YET, then this policy is automatically used for all subcomponents wihtout asking the application.
     * @param initRequ This specifies the number of state query requests that should be sent to the late join RTP/I session before asking someone to join
     *                 the late join RTP/I session via the generic services channel (the separate transport session). It should always be a value larger
     *                 than 3, even when reliable transmission is used. A higher value should be used when the state queries and/or the states are 
     *                 transmitted unreliably. The minimum of 3 is required since there may be brief periods of times when no application is able to 
     *                 answer a state query (for example when the responsibility for relying to state quieries is assigned to the floor holder of a
     *                 subcomponent and that floor is currently being transferred to a new floor holder. The default of this value is 5.
     * @param joinRequ This specifies the number of join requests the late join service will transmit to the generic services channel before ginving up.
     *                 This value defaults to 10. It should only be changed if you know how exactly the mechanism for joining the late join RTP/I session
     *                 works (join implosion avoidance!).
     * @param joinProb This specifies the likeliness that a round of the join implosion avoidance algorithm will be successfull. The default is 0.7.
     *                   Change this value only if you know how the mechanism for joining the late join RTP/I session works (join implosion avoidance).
     */ 


    public LateJoin(LateJoinRecipient ljRecipient,
		    Rtpi r,
		    Rtpi ljR,
		    RtpiSourceInfo ljParticipant,
		    int payloadType,
		    Transport genericServices, 
		    int joinPolicy, 
		    int minMembershipTime, 
		    int maxMembershipTime,
		    long maxStateTransmissionTime,
		    int defaultPolicy,
		    int initRequ,
		    int joinRequ,
		    float joinProb) {

	rtpi=r;
	rtpi.setRtpiRecipient(this);
	ljRtpi=ljR;
	ljRtpi.setRtpiRecipient(this);
	
	recipient = ljRecipient;

	initialRequests = initRequ;
	joinRequests = joinRequ;
	joinProbability = joinProb;

	// setup timer manager
	if (maxStateTransmissionTime < 0) {
	    System.err.println("LateJoin::LateJoin - maxMembershipTime has to be >= 0, set to default 0 ms");
	    maxStateTransmissionTime = 0;
	}
	timerManager = new TimerManager(this, maxStateTransmissionTime);

	// setup subcomponent manager
	if ((defaultPolicy != EVENT) && (defaultPolicy != AT_ONCE) && (defaultPolicy != NOT_INTERESTED) 
	    && (defaultPolicy != NET_CAPACITY) && (defaultPolicy != NO_POLICY_YET)) {
	    System.err.println("LateJoin::LateJoin - unknown defaultPolicy, set to default NO_POLICY_YET");
	    defaultPolicy = NO_POLICY_YET;
	}
	
	subcomponentManager = new SubcomponentManager(this, defaultPolicy);

	// setup client state manager
	clientStateManager = new ClientStateManager();
	clientStateManager.setState(ClientStateManager.CLIENT_COMPLETE);
	clientStateManager.setLearning(true);
	//timerManager.setClientTimeoutTimer();

	// setup server state manager
	if (minMembershipTime <= 0) {
	    System.err.println("LateJoin::LateJoin - minMembershipTime has to be > 0, set to default 1 min");
	    minMembershipTime = 1;
	}
	if (maxMembershipTime <= 0) {
	    System.err.println("LateJoin::LateJoin - maxMembershipTime has to be > 0, set to default 5 min");
	    maxMembershipTime = 5;
	}
	if ((joinPolicy != APPLICATION) && (joinPolicy != AUTOMATIC)) {
	    System.err.println("LateJoin::LateJoin - joinPolicy hast to be AUTOMATIC " +
			       " or APPLICATION, set to default AUTOMATIC");
	    joinPolicy = AUTOMATIC;
	}
        membershipManager = new MembershipManager(this, joinPolicy, minMembershipTime * 60000L, maxMembershipTime * 60000L);
	timerManager.setServerTimeoutTimer(); // in case we are the first member joining


	ljChannelManager = new LateJoinChannelManager(this, rtpi, ljRtpi, ljParticipant, payloadType);

	cChannelManager = new ControlChannelManager(this, genericServices, ljParticipant);

	delayManager = new DelayManager();
	start();
    }


    /**
     * This creates a new late join service instance. Default values are used for the parameters that are responsible
     * for replying to late join state queries and joining the late join session.
     *
     * @param ljRecipient The recipient of information from this instance. Commonly this is the application.
     * @param r The base RTP/I session. The late join service will not transmit anything over this session. However
     *          it needs to monitor incoming RTP/I ADUs and the RTCP/I information.
     * @param ljR The late join RTP/I session. The late join service uses this session to transmit state queries and
     *            states. This RTP/I session is created and destroyed by the application. The application is also responsible
     *            for the QoS settings of the underlying reliability service. The late join service joins and leaves this session
     *            as it is appropriate.
     * @param ljParticipant This information is used for the RTCP/I SDES reporting in the late join session. It should be minimal.
     *                      That is only the CNAME should be set.
     * @param payloadType This is the payload type use by the application.
     * @param genericServices This is the generic services channel. In the current version this is a Transport session which is 
     *                        exclusively used by the late join service to signal that someone should join the late join RTP/I session
     *                        and provide the state of a certain subcomponent. The application providing this Transport object is responsible
     *                        for creating and cleaning up this Transport instance.
     * @param joinPolicy This may be AUTOMATIC if the lateJoin group should be joined automatically using the random number
     *                   mechanisms described above. Or it can be APPLICATION if the application decides for which subcomponents
     *                   this late join should be responsible. The latter should only be chosen if the responsibility for a subcomponent
     *                   rests with exactly one participant in the session.
     * @param minMembershipTime This is the minimal time that a late join service will stay in the late join RTP/I session once it joins.
     *                          At the startup of the late join service it will automatically join the late join RTP/I session. Once it has
     *                          left the late join RTP/I session it may rejoin in order to help a late comer. 
     * @param maxMembershipTime This is the maximal time tat a late join service will stay in the late join RTP/I session once it joins.
     *                          The late join service will choose a value between the minMembershipTime and the maxMembership time for
     *                          leaving the late join session if it has finished its late join for all discovered subcomponents and if it
     *                          has not transmitted any subcomponent state to help another latecomer.
     * @param maxStateTransmissionTime This is the maximum time required by the application to extract and succesfully transmit the state
     *                                 of a subcomponent. Obviously this has to bee an estimate, since the successfull transmission may require
     *                                 retransmissions of lost packets. If this value is too low than the late join service will repeatedly
     *                                 ask for the state of a subcomponent while a member in the late join RTP/I session is already extracting
     *                                 and transmitting the state. If it is set too large, then it takes longer to discover that a state transmission
     *                                 has got lost, or the sender of the state has crashed before it could finish transmitting the state.
     *                                 Therefore this value should be set to a loarge amount if reliable communication is used for the transmission
     *                                 of states over the late join RTP/I session. 
     * @param defaultPolicy One of NO_POLICY_YET, NOT_INTERESTED, AT_ONCE or EVENT. NO_POLICY_YET signals that there is
     *                      no default policy. The meaning of the other policies is explained in the description of the constants. If the defaultPolicy
     *                      is different from NO_POLICY_YET, then this policy is automatically used for all subcomponents wihtout asking the application.
     */ 

    public LateJoin(LateJoinRecipient ljRecipient,
		    Rtpi r,
		    Rtpi ljR,
		    RtpiSourceInfo ljParticipant,
		    int payloadType,
		    Transport genericServices, 
		    int joinPolicy, 
		    int minMembershipTime, 
		    int maxMembershipTime,
		    long maxStateTransmissionTime,
		    int defaultPolicy) {

	this(ljRecipient,r, ljR, ljParticipant, payloadType, genericServices, joinPolicy, 
	     minMembershipTime, maxMembershipTime, maxStateTransmissionTime, 
	     defaultPolicy, INITIAL_REQUESTS_DEFAULT, JOIN_REQUESTS_DEFAULT, 
	     NO_JOIN_PROBABILITY_DEFAULT);
    }

    /**
     * This creates a new late join service instance. Default values are used for the parameters that are responsible
     * for replying to late join state queries and joining the late join session. No default late join policy is specified.
     *
     * @param ljRecipient The recipient of information from this instance. Commonly this is the application.
     * @param r The base RTP/I session. The late join service will not transmit anything over this session. However
     *          it needs to monitor incoming RTP/I ADUs and the RTCP/I information.
     * @param ljR The late join RTP/I session. The late join service uses this session to transmit state queries and
     *            states. This RTP/I session is created and destroyed by the application. The application is also responsible
     *            for the QoS settings of the underlying reliability service. The late join service joins and leaves this session
     *            as it is appropriate.
     * @param ljParticipant This information is used for the RTCP/I SDES reporting in the late join session. It should be minimal.
     *                      That is only the CNAME should be set.
     * @param payloadType This is the payload type use by the application.
     * @param genericServices This is the generic services channel. In the current version this is a Transport session which is 
     *                        exclusively used by the late join service to signal that someone should join the late join RTP/I session
     *                        and provide the state of a certain subcomponent. The application providing this Transport object is responsible
     *                        for creating and cleaning up this Transport instance.
     * @param joinPolicy This may be AUTOMATIC if the lateJoin group should be joined automatically using the random number
     *                   mechanisms described above. Or it can be APPLICATION if the application decides for which subcomponents
     *                   this late join should be responsible. The latter should only be chosen if the responsibility for a subcomponent
     *                   rests with exactly one participant in the session.
     * @param minMembershipTime This is the minimal time that a late join service will stay in the late join RTP/I session once it joins.
     *                          At the startup of the late join service it will automatically join the late join RTP/I session. Once it has
     *                          left the late join RTP/I session it may rejoin in order to help a late comer. 
     * @param maxMembershipTime This is the maximal time tat a late join service will stay in the late join RTP/I session once it joins.
     *                          The late join service will choose a value between the minMembershipTime and the maxMembership time for
     *                          leaving the late join session if it has finished its late join for all discovered subcomponents and if it
     *                          has not transmitted any subcomponent state to help another latecomer.
     * @param maxStateTransmissionTime This is the maximum time required by the application to extract and succesfully transmit the state
     *                                 of a subcomponent. Obviously this has to bee an estimate, since the successfull transmission may require
     *                                 retransmissions of lost packets. If this value is too low than the late join service will repeatedly
     *                                 ask for the state of a subcomponent while a member in the late join RTP/I session is already extracting
     *                                 and transmitting the state. If it is set too large, then it takes longer to discover that a state transmission
     *                                 has got lost, or the sender of the state has crashed before it could finish transmitting the state.
     *                                 Therefore this value should be set to a loarge amount if reliable communication is used for the transmission
     *                                 of states over the late join RTP/I session. 
     */ 

    public LateJoin(LateJoinRecipient ljRecipient,
		    Rtpi r,
		    Rtpi ljR,
		    RtpiSourceInfo ljParticipant,
		    int payloadType,
		    Transport genericServices, 
		    int joinPolicy, 
		    int minMembershipTime, 
		    int maxMembershipTime,
		    long maxStateTransmissionTime) {

	this(ljRecipient,r, ljR, ljParticipant, payloadType, genericServices, joinPolicy, 
	     minMembershipTime, maxMembershipTime, maxStateTransmissionTime, 
	     NO_POLICY_YET, INITIAL_REQUESTS_DEFAULT, JOIN_REQUESTS_DEFAULT, 
	     NO_JOIN_PROBABILITY_DEFAULT);
    }

    LateJoinRecipient getLateJoinRecipient() {
	return recipient;
    }

    LateJoinChannelManager getLJChannelManager() {
	return ljChannelManager;
    }

    ControlChannelManager getControlChannelManager() {
	return cChannelManager;
    }

    SubcomponentManager getSubcomponentManager() {
	return subcomponentManager;
    }

    TimerManager getTimerManager() {
	return timerManager;
    }

    ClientStateManager getClientStateManager() {
	return clientStateManager;
    }

    MembershipManager getMembershipManager() {
	return membershipManager;
    }

    DelayManager getDelayManager() {
	return delayManager;
    }

    public void run() {
	LateJoinMessage message = null;
	while (true) {
	    try {
		message = (LateJoinMessage)messages.get();
	    } catch (Exception ex) {
		System.err.println("LateJoin::run - could not get next message!");
		continue;
	    }
	    switch(message.opCode) {
	    case LateJoinMessage.QUIT: {
		execQuit();
		return;
	    }
	    case LateJoinMessage.RECEIVE_EVENT: {
		PacketMessageArgs args = (PacketMessageArgs)message.args;
		execReceiveEvent(args.rtpi, (RtpiEvent)args.data);
		break;
	    }
	    case LateJoinMessage.RECEIVE_STATE: {
		PacketMessageArgs args = (PacketMessageArgs)message.args;
		execReceiveState(args.rtpi, (RtpiState)args.data);
		break;
	    }
	    case LateJoinMessage.RECEIVE_DELTA_STATE: {
		PacketMessageArgs args = (PacketMessageArgs)message.args;
		execReceiveDeltaState(args.rtpi, (RtpiDeltaState)args.data);
		break;
	    }
	    case LateJoinMessage.RECEIVE_STATE_QUERY: {
		PacketMessageArgs args = (PacketMessageArgs)message.args;
		execReceiveStateQuery(args.rtpi, (RtpiStateQuery)args.data);
		break;
	    }
	    case LateJoinMessage.ADD_SUBCOMPONENT: {
		AddSubcomponentArgs args = (AddSubcomponentArgs) message.args;
		execAddSubcomponent(args.rtpi, args.id, args.active, args.name);
		break;
	    }
	    case LateJoinMessage.REMOVE_SUBCOMPONENT: {
		SubcomponentArgs args = (SubcomponentArgs) message.args;
		execRemoveSubcomponent(args.rtpi, args.subID);
		break;
	    }
	    case LateJoinMessage.ACTIVATE_SUBCOMPONENT: {
		SubcomponentArgs args = (SubcomponentArgs) message.args;
		execActivateSubcomponent(args.rtpi, args.subID);
		break;
	    }
	    case LateJoinMessage.DEACTIVATE_SUBCOMPONENT: {
		SubcomponentArgs args = (SubcomponentArgs) message.args;
		execDeactivateSubcomponent(args.rtpi, args.subID);
		break;
	    }
	    case LateJoinMessage.CHANGE_SOURCE_INFO: {
		ParticipantMessageArgs args = (ParticipantMessageArgs)message.args;
		execChangeSourceInfo(args.rtpi, args.participant);
		break;
	    }
	    case LateJoinMessage.REMOVE_SOURCE: {
		ParticipantMessageArgs args = (ParticipantMessageArgs)message.args;
		execRemoveSource(args.rtpi, args.participant);
		break;
	    }
	    case LateJoinMessage.SET_POLICY: {
		PolicyMessageArgs args = (PolicyMessageArgs)message.args;
		execSetPolicy(args.id, args.policy);
		break;
	    }
	    case LateJoinMessage.RECEIVE_CONTROL_PACKET: {
		cChannelManager.execReceiveTransportPacket((TransportPacket)message.args);
		break;
	    }
	    case LateJoinMessage.ACKNOWLEDGE_TRANSMIT_STATE: {
		execAcknowledgeTransmitState(((Long)message.args).longValue());
		break;
	    }
	    case LateJoinMessage.LJ_CHANNEL_TRANSMIT_STATE: {
		ljChannelManager.transmitState((RtpiState)message.args);
		break;
	    }
	    case LateJoinMessage.TIMER_STATE_QUERY: {
		timerManager.handleStateQueryTimer((StateTimerObject)message.args);
		break;
	    }
	    case LateJoinMessage.TIMER_STATE_REPLY: {
		timerManager.handleStateReplyTimer((StateTimerObject)message.args);
		break;
	    }
	    case LateJoinMessage.TIMER_NET_QUERY: {
		timerManager.handleNetQueryTimer((TimerObject)message.args);
		break;
	    }
	    case LateJoinMessage.TIMER_SERVER_TIMEOUT: {
		timerManager.handleServerTimeoutTimer((TimerObject)message.args);
		break;
	    }
	    case LateJoinMessage.TIMER_CLIENT_TIMEOUT: {
		timerManager.handleClientTimeoutTimer((TimerObject)message.args);
		break;
	    }
	    case LateJoinMessage.LOCAL_SUBCOMPONENT_ADDED: {
		execAddSubcomponent(((Long)message.args).longValue());
		break;
	    }
	    case LateJoinMessage.REGISTER_SUBCOMPONENT: {
		subcomponentManager.registerSubcomponent(((Long)message.args).longValue());
		break;
	    }
	    case LateJoinMessage.UNREGISTER_SUBCOMPONENT: {
		subcomponentManager.unregisterSubcomponent(((Long)message.args).longValue());
		break;
	    }
	    case LateJoinMessage.RTPI_ADU_LOST: {
		RtpiAduLostArgs args = (RtpiAduLostArgs) message.args;
		execRtpiAduLost(args.r, args.participantID, args.subID, args.type, args.sequenceNumber, args.timestamp);
		break;
	    }
	    case LateJoinMessage.RTPI_COULD_NOT_RECOVER: {
		RtpiAduLostArgs args = (RtpiAduLostArgs) message.args;
		execRtpiCouldNotRecover(args.r, args.participantID, args.subID, args.type, args.sequenceNumber, args.timestamp);
		break;
	    }
	    case LateJoinMessage.CONNECTION_CLOSED: {
		execConnectionClosed((Rtpi) message.args);
	    }
	    }
	}
    }

    SyncQueue messages() {
	return messages;
    }

    /**
     * This terminates the late join service. The application is resposible for quitting
     * the RTP/I session and the generic services Transport instance.
     */

    public void quit() {
	messages.put(new LateJoinMessage(LateJoinMessage.QUIT,null));
    }

    private void execQuit() {
	try {
	    if (membershipManager.getState() == MembershipManager.MEMBER)
		ljChannelManager.leaveGroup();
	} catch (Exception ex) {
	    System.err.println("LateJoin: quit: error while leaving the late join group!");
	}
	try {
	    cChannelManager.quit();
	} catch (Exception ex) {
	    System.err.println("LateJoin: quit: error while quitting the control channel manager!");
	}
	try {
	    timerManager.quit();
	} catch (Exception ex) {
	    System.err.println("LateJoin: quit: error while quitting the control channel manager!");
	}	
    }

    /**
     * This is used by the application to set the late join policy for a given subcomponent.
     * It is not permitted to set the policy of a subcomponent that has not yet been discovered
     * by the late join service. This method may be used to reset the late join policy for a
     * subcomponent. In this way the late join service can be abused to fetch the state of 
     * a subcomponent in a scalable and late join policy driven way.
     *
     * @param id The subcomponent ID of the subcomponent for which the policy should be set.
     * @param policy The new late join policy (see class Policy for possible values).
     */

    public void setPolicy(long id, int policy) {
	messages.put(new LateJoinMessage(LateJoinMessage.SET_POLICY,
					 new PolicyMessageArgs(id, policy)));
    }

    private void execSetPolicy(long id, int policy) {
	Subcomponent sub = subcomponentManager.getSubcomponent(id);

	if (sub == null) {
	    System.err.println("LateJoin::setPolicy - Subcomponent with id " +
			       id + " does not exist");
	    return;
	}

	if (sub.getPolicy() != NO_POLICY_YET) { // Policy is reset 
	    timerManager.deleteServerTimeoutTimer();
	    clientStateManager.setState(ClientStateManager.CLIENT_INCOMPLETE);
	    if (membershipManager.getState() == MembershipManager.NO_MEMBER){
		ljChannelManager.joinGroup();
	    }
	} else {
	    // in case the state was faster than our SyncQueue
	    if (sub.getState() == Subcomponent.COMPLETE) return;
	}

	sub.setState(Subcomponent.POLICY_SET);
	sub.setPolicy(policy);

	if (policy == NOT_INTERESTED) {
	    sub.setState(Subcomponent.NOT_INTERESTED);
	    subcomponentManager.checkClientState();  // client complete ?
	    return;
	}
	if (policy == AT_ONCE) {
	    timerManager.setStateQueryTimer(id);
	    sub.setState(Subcomponent.REQUESTED);
	    return;
	}	    
	if (policy == NET_CAPACITY) {
	    // ggf net query timer stellen
	    subcomponentManager.addNetCapacitySubcomponent(id);
	    return;
	}
    }

    /**
     * This method is called by the application to signal that it has
     * created a new subcomponent. It is assumed that the application 
     * does have a valid state of the subcomponent it has created.
     *
     * @param id The subcomponent ID of the subcomponent.
     */

    public void addSubcomponent(long id) {
	messages.put(new LateJoinMessage(LateJoinMessage.LOCAL_SUBCOMPONENT_ADDED,
					 new Long(id)));
    }

    private void execAddSubcomponent(long id) {
	subcomponentManager.addLocalSubcomponent(id);
    }

    /**
     * This method is called by the application when it has been notified that the late
     * join service needs the state of a certain subcomponent and the application is
     * able and willing to extract the state of the subcomponent. As soon as this method
     * is called the late join service will notify the other session participants that
     * it will be able to deliver the state soon. This will suppress further requests
     * for this subcomponent for some time. This mechanism is required since extracting
     * the state of a subcomponent may be a process that consumes some time. Therefore the
     * late join service needs to know whether the application will eventually deliver the
     * state of the subcomponent.
     *
     * @param id The subcomponent ID of the subcomponent.
     */

    public void acknowledgeTransmitState(long id) {
	messages.put(new LateJoinMessage(LateJoinMessage.ACKNOWLEDGE_TRANSMIT_STATE,
					 new Long(id)));
    }

    private void execAcknowledgeTransmitState(long id) {
	Subcomponent sub = subcomponentManager.getSubcomponent(id);
	sub.setLastTimeSeenNow();
	sub.setAcknowledged(true);

	cChannelManager.acknowledgeServerSelection(id);
    }

    /**
     * This is called by the application when it has been requested by the late join servie
     * to extract the state of a certian subcomponent in order to help a late comer. Note 
     * that the application should call acknowledgeTrasnmitState immediately when it is
     * able and willing to provide the state of the subcomponent. The transmitState method
     * is called to deliver the state to the late join service. It is OK for the application
     * to need some time to extract the state. The state will be transmitted over the late
     * join RTP/I session. It is therefore important that the sequence number of the state
     * is not taken from the sequence number space of the regular RTP/I session.
     *
     * @param state The state of the subcomponent.
     */

    public void transmitState(RtpiState state) {
	messages.put(new LateJoinMessage(LateJoinMessage.LJ_CHANNEL_TRANSMIT_STATE,
					 state));
    }

    /**
     * This is called by the application to indicate that it is responsible for the management of
     * a certain subcomponents state. This is required in the case that the join policy APPLICATION
     * has been chosen when the late join instance was created. When the state of this subcomponent
     * is requested, then this late join instance knows that the local application will provide the
     * state of the subcomponent. Therefore there needs to be no suppression mechanism for the replies
     * to state queries or joining the late join session.
     *
     * @param id The subcomponentID of the subcomponent.
     */

    public void claimSubcomponentResponsibility(long id) {
	messages.put(new LateJoinMessage(LateJoinMessage.REGISTER_SUBCOMPONENT,
					 new Long(id)));
    }

    /**
     * This is called by the application to indicate that it is no longer responsible for the management of
     * a certain subcomponents state. This is required in the case that the join policy APPLICATION
     * has been chosen when the late join instance was created. 
     *
     * @param id The subcomponentID of the subcomponent.
     */

    public void releaseSubcomponentResponsibility(long id) {
	messages.put(new LateJoinMessage(LateJoinMessage.UNREGISTER_SUBCOMPONENT,
					 new Long(id)));
    }

    /**
     * This is called by the RTP/I instances when an event has been received.
     *
     * @param r The RTP/I instance that received the event.
     * @param event The event.
     */

    public void receiveEvent(Rtpi r, RtpiEvent event) {
	if (r!=rtpi) {
	    System.err.println("LateJoin: received RtpiEvent for wrong rtpi session - ignoring event!");
	    return;
	}
	messages.put(new LateJoinMessage(LateJoinMessage.RECEIVE_EVENT,
					 new PacketMessageArgs(r, event)));
    }

    private void execReceiveEvent(Rtpi r, RtpiEvent event) {
	recipient.receiveEvent(r, event);

	delayManager.addDelayValue(event.getParticipantID(), event.getTimestamp());

	if (clientStateManager.getState() == ClientStateManager.CLIENT_INCOMPLETE)
	    ljChannelManager.receiveEvent(event.getSubcomponentID());
    }

    /**
     * This is called by the RTP/I instances when a state has been received.
     *
     * @param r The RTP/I instance that received the state.
     * @param state The state.
     */

    public void receiveState(Rtpi r, RtpiState state) {
	messages.put(new LateJoinMessage(LateJoinMessage.RECEIVE_STATE,
					    new PacketMessageArgs(r, state)));

    }

    private void execReceiveState(Rtpi r, RtpiState state) {
	if (r==rtpi) {
	    recipient.receiveState(r, state);
	} else {
	    recipient.lateJoinReceiveState(r, state);
	}

	delayManager.addDelayValue(state.getParticipantID(),state.getTimestamp());

	//if (clientStateManager.getState() == ClientStateManager.CLIENT_INCOMPLETE)
	ljChannelManager.receiveSubcomponentState(state.getSubcomponentID());
    }

    /**
     * This is called by the RTP/I instances when a delta state has been received.
     *
     * @param r The RTP/I instance that received the delta state.
     * @param state The delta state.
     */

    public void receiveDeltaState(Rtpi r, RtpiDeltaState state) {
	if (r!=rtpi) {
	    System.err.println("LateJoin: received RtpiDeltaState for wrong rtpi session - ignoring delta state!");
	    return;
	}
	messages.put(new LateJoinMessage(LateJoinMessage.RECEIVE_DELTA_STATE,
					 new PacketMessageArgs(r, state)));

    }

    private void execReceiveDeltaState(Rtpi r, RtpiDeltaState state) {
	recipient.receiveDeltaState(r, state);

	delayManager.addDelayValue(state.getParticipantID(), state.getTimestamp());

	if (clientStateManager.getState() == ClientStateManager.CLIENT_INCOMPLETE)
	    ljChannelManager.receiveEvent(state.getSubcomponentID());  // same procedure
    }

    /**
     * This is called by the RTP/I instances when a state query has been received.
     *
     * @param r The RTP/I instance that received the state query.
     * @param query The state query.
     */

    public void receiveStateQuery(Rtpi r, RtpiStateQuery query) {
	messages.put(new LateJoinMessage(LateJoinMessage.RECEIVE_STATE_QUERY,
					 new PacketMessageArgs(r, query)));

    }

    private void execReceiveStateQuery(Rtpi r, RtpiStateQuery query) {
	if (r == ljRtpi) {
	    ljChannelManager.receiveStateQuery(query);
	    return;
	}

	// application's channel
	recipient.receiveStateQuery(r, query);
    }

    /**
     * This is called by an RTP/I instance when a new subcomponent has been discovered via
     * an RTCP/I SUBREP packet.
     *
     * @param r The RTP/I instance that discovered the new subcomponent.
     * @param id The subcomponent ID of the discovered subcomponent.
     * @param active A flag indicating whether the discovered subcomponent was reported as active.
     * @param name The application level name (if any) of the subcomponent.
     */
  
    public void addSubcomponent(Rtpi r, long id, boolean active, byte[] name) {
	if (r!=rtpi) {
	    System.err.println("LateJoin: add subcomponent for wrong rtpi session - ignoring add subcomponent!");
	    return;
	}
	messages.put(new LateJoinMessage(LateJoinMessage.ADD_SUBCOMPONENT,
					 new AddSubcomponentArgs(r, id, name, active)));

    }

    private void execAddSubcomponent(Rtpi r, long id, boolean active, byte[] name) {
	recipient.addSubcomponent(r, id, active, name);

	subcomponentManager.addRemoteSubcomponent(id);

	if (clientStateManager.getLearning()) { // active client phase completed ?
	    Subcomponent ljSub = subcomponentManager.getSubcomponent(id);
	    if (subcomponentManager.useDefaultPolicy()) {
		int policy = subcomponentManager.getDefaultPolicy();
		ljSub.setPolicy(policy);
		if (policy == AT_ONCE) {
		    timerManager.setStateQueryTimer(id);
		    ljSub.setState(Subcomponent.REQUESTED);
		}
	    } else {
		recipient.lateJoinAskForPolicy(id);
	    }
	}
    }
    

    /**
     * This is called by an RTP/I instance when a subcomponent has timed out.
     *
     * @param r The RTP/I instance for which the subcomponent did time out.
     * @param id The subcomponent ID of the timed out subcomponent.
     */
  
    public void removeSubcomponent(Rtpi r, long id) {
	if (r!=rtpi) {
	    System.err.println("LateJoin: remove subcomponent for wrong rtpi session - ignoring remove subcomponent!");
	    return;
	}

	messages.put(new LateJoinMessage(LateJoinMessage.REMOVE_SUBCOMPONENT,
					 new SubcomponentArgs(r,id)));

    }

    private void execRemoveSubcomponent(Rtpi r, long id) {
	recipient.removeSubcomponent(r, id);
    }


    /**
     * This is called by an RTP/I instance when a subcomponent has become active for a remote participant.
     *
     * @param r The RTP/I instance for which the subcomponent became active.
     * @param id The subcomponent ID of the activated subcomponent.
     */

    public void activateSubcomponent(Rtpi r, long id) {
	if (r!=rtpi) {
	    System.err.println("LateJoin: activate subcomponent for wrong rtpi session - ignoring activate subcomponent!");
	    return;
	}
	messages.put(new LateJoinMessage(LateJoinMessage.ACTIVATE_SUBCOMPONENT,
					 new SubcomponentArgs(r,id)));

    }

    private void execActivateSubcomponent(Rtpi r, long id) {
	recipient.activateSubcomponent(r, id);
    }


    /**
     * This is called by an RTP/I instance when a subcomponent has been deactivated for the session.
     *
     * @param r The RTP/I instance for which the subcomponent was deactivated.
     * @param id The subcomponent ID of the deactivated subcomponent.
     */

    public void deactivateSubcomponent(Rtpi r, long id) {
	if (r!=rtpi) {
	    System.err.println("LateJoin: deactivate subcomponent for wrong rtpi session - ignoring deactivate subcomponent!");
	    return;
	}

	messages.put(new LateJoinMessage(LateJoinMessage.DEACTIVATE_SUBCOMPONENT,
					 new SubcomponentArgs(r,id)));
    }

    private void execDeactivateSubcomponent(Rtpi r, long id) {
	recipient.deactivateSubcomponent(r,id);
    }

    /**
     * This is called by an RTP/I instance when the information about a remote source has changed via an RTCP/I
     * SDES packet.
     *
     * @param r The RTP/I instance for which the information changed.
     * @param participant The new information about the source.
     */

    public void changeSourceInfo(Rtpi r, RtpiSourceInfo participant) {
	messages.put(new LateJoinMessage(LateJoinMessage.CHANGE_SOURCE_INFO,
					 new ParticipantMessageArgs(r, participant)));
    }

    private void execChangeSourceInfo(Rtpi r, RtpiSourceInfo participant) {
	if (r == rtpi) {
	    recipient.changeSourceInfo(r, participant);
	    return;
	}
	
	if (!delayManager.participantExists(participant.getParticipantID())) {
	    membershipManager.incrNbrOfMembers();
	    delayManager.addParticipant(participant.getParticipantID());
	}
    }

    /**
     * This is called by an RTP/I instance when a remote source has timed out.
     *
     * @param r The RTP/I instance for which the source timed out.
     * @param participant The source that timed out.
     */

    public void removeSource(Rtpi r, RtpiSourceInfo participant) {
	messages.put(new LateJoinMessage(LateJoinMessage.REMOVE_SOURCE,
					 new ParticipantMessageArgs(r, participant)));
    }

    private void execRemoveSource(Rtpi r, RtpiSourceInfo participant) {
	if (r==rtpi) {
	    recipient.removeSource(r, participant);
	    return;
	}

 	if (delayManager.participantExists(participant.getParticipantID())) {
	    membershipManager.decrNbrOfMembers();
	    delayManager.removeParticipant(participant.getParticipantID());
	}
    }

    /**
     * This is called by an RTP/I instance when a packet got lost.
     * If the RTP/I instance is the base RTP/I session, then the message will be simply forwarded to the application.
     * However, when it concerns a state that has been transmitted over the late join RTP/I session and this late join
     * instance needed this state, it is examined more closely. When the late join service has already informed the application
     * about the failiure of the late join for the subcomponent, then this message is ignored. However, when the late join
     * service is still in the process of transmitting state queries, then the state query process is started from scratch.
     *
     * @param r The RTP/I session that experienced the loss.
     * @param participantID The ID of the sender of the lost ADU.
     * @param subID The subcomponent ID that this ADU refers to.
     * @param type The type of the ADU (e.g. state, event, etc.).
     * @param sequenceNumber The most recent sequencenumber that got lost.
     * @param timestamp The timestamp of the most recent ADU that got lost.
     */

    public void rtpiAduLost(Rtpi r, int participantID, long subID, int type, int sequenceNumber, int timestamp) {
	messages.put(new LateJoinMessage(LateJoinMessage.RTPI_ADU_LOST,
					 new RtpiAduLostArgs(r, participantID, subID, type, sequenceNumber,timestamp)));
    }

    private void execRtpiAduLost(Rtpi r, int participantID, long subID, int type, int sequenceNumber, int timestamp) {
	if (r==rtpi) {
	    recipient.rtpiAduLost(r,participantID,subID,type,sequenceNumber,timestamp);
	    return;
	}
	timerManager.resetClientRequestCounter(subID);
    }


    /**
     * This is called by an RTP/I instance when it could not recover a lost packet, even though it was instructed to
     * do so. If the RTP/I instance is the base RTP/I session, then the message will be simply forwarded to the application.
     * However, when it concerns a state that has been transmitted over the late join RTP/I session and this late join
     * instance needed this state, it is examined more closely. When the late join service has already informed the application
     * about the failiure of the late join for the subcomponent, then this message is ignored. However, when the late join
     * service is still in the process of transmitting state queries, then the state query process is started from scratch.
     *
     * @param r The RTP/I session that experienced the loss.
     * @param participantID The ID of the sender of the lost ADU.
     * @param subID The subcomponent ID that this ADU refers to.
     * @param type The type of the ADU (e.g. state, event, etc.).
     * @param sequenceNumber The most recent sequencenumber that got lost.
     * @param timestamp The timestamp of the most recent ADU that got lost.
     */

    public void rtpiCouldNotRecover(Rtpi r, int participantID, long subID, int type, int sequenceNumber, int timestamp) {
	messages.put(new LateJoinMessage(LateJoinMessage.RTPI_COULD_NOT_RECOVER,
					 new RtpiAduLostArgs(r, participantID, subID, type, sequenceNumber,timestamp)));
    }

    private void execRtpiCouldNotRecover(Rtpi r, int participantID, long subID, int type, int sequenceNumber, int timestamp) {
	if (r==rtpi) {
	    recipient.rtpiCouldNotRecover(r,participantID,subID,type,sequenceNumber,timestamp);
	    return;
	}
	timerManager.resetClientRequestCounter(subID);
    }
	    
    /**
     * This is called when an RTP/I session dies unexpectantly. It causes the shutdown of the late join
     * service. The application is informed about this. The application either receives a lateJoineTerminated
     * message if the late join RTP/I session failed or a connectionClosed message followed by a lateJoinTerminated
     * message when the main RTP/I session died.
     *
     * @param r The RTP/I session that died.
     */
   
    public void connectionClosed(Rtpi r){
	messages.put(new LateJoinMessage(LateJoinMessage.CONNECTION_CLOSED, r));
    }
    
    private void execConnectionClosed(Rtpi r) {
	if (r==rtpi) {
	    recipient.connectionClosed(r);
	    recipient.lateJoinTerminated();
	} else {
	    recipient.lateJoinTerminated();
	}
	quit();
    }
}









