/* The Java RTP/I latejoin service, Version 0.1 alpha.
 *  
 * This library provides the functionality of a generic RTP/I latejoin service.
 *
 * Copyright (C) 2000 Juergen Vogel, Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Juergen Vogel
 * Martin Mauve
 *
 * e-mail:
 * {vogel,mauve}@informatik.uni-mannheim.de
 * 
 * paper mail:
 *
 * Juergen Vogel
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.services.latejoin;

import java.util.Hashtable;

class MembershipManager {
    private static boolean debug = false;

    static final int MEMBER    = 0;
    static final int NO_MEMBER = 1;

    private LateJoin lj;

    private int state;
    private int policy;

    // variables needed to compute timeout-interval
    private int nbrOfServableQueries;
    private int nbrOfAnsweredQueries;
    private long maxTime;
    private long minTime;

    // variables needed to decide on joining lj-mc-group
    private int nbrOfMembers;

    MembershipManager(LateJoin l, int p, long min, long max) {
	lj = l;
	policy = p;
	nbrOfServableQueries = 0;
	nbrOfAnsweredQueries = 0;
	nbrOfMembers = 1; //ourselves !!
	minTime = min;
	maxTime = max;
	state=NO_MEMBER;
    }

    int getState() {
	return state;
    }

    void setState(int s) {
	state = s;
    }

    int getPolicy() {
	return policy;
    }

    void reset() {
	nbrOfServableQueries = 0;
	nbrOfAnsweredQueries = 0;
    }

    void incrNbrOfServableQueries() {
	nbrOfServableQueries++;
    }

    void incrNbrOfAnsweredQueries() {
	nbrOfAnsweredQueries++;
    }

    void incrNbrOfMembers() {
	nbrOfMembers++;

	if (debug)
	    System.out.println("MembershipManager::incrNbrOfMembers - " + nbrOfMembers);
    }

    void decrNbrOfMembers() {
	nbrOfMembers--;

	if (debug)
	    System.out.println("MembershipManager::decrNbrOfMembers - " + nbrOfMembers);
    }

    long computeTimeout() {
	SubcomponentManager sm = lj.getSubcomponentManager();

	if (debug)
	    System.out.println("MembershipManager::computeTimeout - nbr answered/servable queries: " +
			       nbrOfAnsweredQueries + "/" + nbrOfServableQueries +
			       " nbr possessed/total subcomponents: " +
			       sm.getNbrOfPossessedSubcomponents() + "/" + sm.getNbrOfSubcomponents());

	if ((nbrOfServableQueries == 0) || (sm.getNbrOfSubcomponents() == 0) )
	    return minTime;

	double factor1 = (double)nbrOfAnsweredQueries/(double)nbrOfServableQueries;
	double factor2 = (double)sm.getNbrOfPossessedSubcomponents()/(double)sm.getNbrOfSubcomponents();
	double time = ((double)maxTime/2)*(factor1 + factor2);
	long timeout = Math.round(time);

	if (timeout < minTime)
	    return minTime;
	else
	    return timeout;
    }

    boolean joinLJGroup(int tries, int ssrc) {
	DelayManager dm = lj.getDelayManager();

	//1. calculate probability
	if (nbrOfMembers == 2) return true;

	double p = 1.0 - Math.pow(lj.joinProbability, 1.0 / (double)(nbrOfMembers - 1));
	p *= ((tries + 2) / 3);
	p *= ((double)dm.getAverageDelay() / (double)dm.getDelay(ssrc));

	if (p >= 1.0) return true;

	//2. check if we join
	double r = Math.random();

	if (debug)
	    System.out.println("MembershipManager::joinLJGroup - joining probability: " + p +
			       " random variable: " + r);

	if (r <= p) // r E [0;p]
	    return true;
	return false;
    }
}











