/* The Java RTP/I latejoin service, Version 0.1 alpha.
 *  
 * This library provides the functionality of a generic RTP/I latejoin service.
 *
 * Copyright (C) 2000 Juergen Vogel, Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Juergen Vogel
 * Martin Mauve
 *
 * e-mail:
 * {vogel,mauve}@informatik.uni-mannheim.de
 * 
 * paper mail:
 *
 * Juergen Vogel
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.services.latejoin;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

class SubcomponentManager {
    private boolean debug = false;

    private static long SUPRESSION_TIME = 5000;

    private LateJoin lj;
    private Hashtable subcomponents;
    private boolean useDefPolicy;
    private Vector netCapSubcomponents;

    private int defaultPolicy;

    SubcomponentManager(LateJoin l, int defPol) {
	lj = l;
	subcomponents = new Hashtable();
	netCapSubcomponents = new Vector();
	defaultPolicy=defPol;	
	useDefPolicy = !(defaultPolicy==LateJoin.NO_POLICY_YET);
    }

    int getDefaultPolicy() {
	return defaultPolicy;
    }

    int getNbrOfSubcomponents() {
	return subcomponents.size();
    }

    int getNbrOfPossessedSubcomponents() {
	int nbr = 0;
	for (Enumeration e = subcomponents.elements(); e.hasMoreElements(); )
	    if (((Subcomponent)e.nextElement()).getState() == Subcomponent.COMPLETE)
		nbr++;

	return nbr;
    }

    boolean useDefaultPolicy() {
	return useDefPolicy;
    }

    void registerSubcomponent(long id) { //set application responsibility for this subcomp
	Subcomponent sub = (Subcomponent)subcomponents.get(new Long(id));

	if (sub == null) {
	    System.err.println("SubcomponentManager::registerSubcomponent - Subcomponent " +
			       "with ID " + id + " does not exist");
	    return;
	}

	sub.setRegistered(true);
    }

    void unregisterSubcomponent(long id) {
	Subcomponent sub = (Subcomponent)subcomponents.get(new Long(id));

	if (sub == null) {
	    System.err.println("SubcomponentManager::unregisterSubcomponent - Subcomponent " +
			       "with ID " + id + " does not exist");
	    return;
	}

	sub.setRegistered(false);
    }

    void addRemoteSubcomponent(long id) {
	if (subcomponents.get(new Long(id)) != null) return;  // already exists

	ClientStateManager cm = lj.getClientStateManager();

	Subcomponent ljSub = new Subcomponent();
	ljSub.setState(Subcomponent.SUBCOMP_DISCOVERED);
	subcomponents.put(new Long(id), ljSub);

	// the discovered subcomponent does not concern the late join -> it is de facto new
	if (cm.getLearning() == false) {
	    if (debug) {
		System.out.println("SubcomponentManager::addRemoteSubcomponent - "
				   + "new Subcomponent with ID " + id + " but learning phase is completed");
	    }
	    return;
	}

	cm.setState(ClientStateManager.CLIENT_INCOMPLETE);
	lj.getTimerManager().deleteServerTimeoutTimer();

	if (debug) {
	    System.out.println("SubcomponentManager::addRemoteSubcomponent - "
			       + "new Subcomponent with ID " + id);
	}
    }

    void addLocalSubcomponent(long id) {
	if (subcomponents.get(new Long(id)) != null) return;  // already exists

	Subcomponent ljSub = new Subcomponent();
	ljSub.setState(Subcomponent.COMPLETE);
	subcomponents.put(new Long(id), ljSub);

	if (debug) {
	    System.out.println("SubcomponentManager::addLocalSubcomponent - "
			       + "new Subcomponent with ID " + id);
	}
     }

    Subcomponent getSubcomponent(long id) {
	return (Subcomponent)subcomponents.get(new Long(id));
    }

    void stateQueryFailed(long id) {
	Subcomponent sub = (Subcomponent)subcomponents.get(new Long(id));
	if (sub == null) {
	    System.err.println("SubcomponentManager::stateQueryFailed - Subcomponent " +
			       "with ID " + id + " does not exist");
	    return;
	}

	sub.setState(Subcomponent.FAILED);

	lj.getLateJoinRecipient().lateJoinFailed(id);

	if (debug)
	    System.out.println("SubcomponentManager::stateQueryFailed - " +
			       "request failed for Subcomponent with ID " + id);
    }

    boolean checkSubcomponentPossession(long id) {
	Subcomponent sub = (Subcomponent)subcomponents.get(new Long(id));
	if (sub == null) return false;
	if (sub.getState() == Subcomponent.COMPLETE) return true;
	return false;
    }

    boolean checkSubcomponentRegistration(long id) {
	Subcomponent sub = (Subcomponent)subcomponents.get(new Long(id));
	if (sub == null) return false;
	return sub.getRegistered();
    }

    boolean checkLastTimeSeen(long id) {
	Subcomponent sub = (Subcomponent)subcomponents.get(new Long(id));
	if (sub == null) return false;

	DelayManager dm = lj.getDelayManager();
	long lts = sub.getLastTimeSeen();

	// never received the state before
	if (lts == -1) {
	    if (debug)
		System.out.println("SubcomponentManager::checkLatTimeSeen - " +
				   "the state was not transmitted once");
	    return true;
	}

	// haven't received the state since a long time  -> SRM
	// if ((System.currentTimeMillis() - lts) > (dm.getAverageDelay()*2)) {
	if ((System.currentTimeMillis() - lts) > SUPRESSION_TIME) {
	    if (debug)
		System.out.println("SubcomponentManager::checkLastTimeSeen - " +
				   "the last state transmission was long ago");
	    return true;
	}

	// there has been a transmission recently
	if (debug)
	    System.out.println("SubcomponentManager::checkLastTimeSeen - " +
			       "don't answer because of recent transmission");
	return false;
    }

    void handleStateQuery(long id, int ssrc) {
	// version for lj-mc-group
	TimerManager tm = lj.getTimerManager();
	MembershipManager mm = lj.getMembershipManager();

	if (tm.stateReplyTimerRunning(id)) // reply timer already running
	    return;

	if (getSubcomponent(id).getAcknowledged()) { // we are transmitting the state soon
	    lj.getControlChannelManager().acknowledgeServerSelection(id);
	    return;
	}

	// are we client as well?
	if (tm.stateQueryTimerRunning(id)) { // avoid state query implosion
	    tm.resetStateQueryTimer(id);
	    return;
	}

	switch(mm.getPolicy()) {
	case LateJoin.APPLICATION:
	    {
		if (checkSubcomponentRegistration(id) && checkLastTimeSeen(id))
		    // answer at once
		    lj.getLateJoinRecipient().lateJoinGetState(id);
		    getSubcomponent(id).setLastTimeSeenNow();
		break;
	    }
	case LateJoin.AUTOMATIC:
	    {
		if (checkSubcomponentPossession(id) && checkLastTimeSeen(id)) {
		    mm.incrNbrOfServableQueries();
		    tm.setStateReplyTimer(id, ssrc);
		}
		break;
	    }
	}
    }


    void handleStateQuery(long id, int ssrc, int tries) {
	// version for state queries received on the data group control channel for generic services
	TimerManager tm = lj.getTimerManager();
	MembershipManager mm = lj.getMembershipManager();

	if (tm.stateReplyTimerRunning(id)) // reply timer already running
	    return;

	if (getSubcomponent(id).getAcknowledged()) { // we are transmitting the state soon
	    lj.getControlChannelManager().acknowledgeServerSelection(id);
	    return;
	}

	// are we client as well?
	if (tm.stateQueryTimerRunning(id)) { // avoid state query implosion
	    tm.resetStateQueryTimer(id);
	    return;
	}

	if (mm.getState() == MembershipManager.MEMBER) { // we ARE member, use other version
	    handleStateQuery(id, ssrc);
	    return;
	}

	switch(mm.getPolicy()) {
	case LateJoin.APPLICATION:
	    {
		if (checkSubcomponentRegistration(id) && checkLastTimeSeen(id)) {
		    lj.getLJChannelManager().joinGroup();
		    lj.getLateJoinRecipient().lateJoinGetState(id);
		    getSubcomponent(id).setLastTimeSeenNow();
		}
		break;
	    }
	case LateJoin.AUTOMATIC:
	    {
		if (checkSubcomponentPossession(id) && checkLastTimeSeen(id)) {
		    if (mm.joinLJGroup(tries, ssrc)) {
			tm.setServerTimeoutTimer();
			lj.getLJChannelManager().joinGroup();
			mm.incrNbrOfServableQueries();
			tm.setStateReplyTimer(id, ssrc);
		    } else {
			if (debug)
			    System.out.println("SubcomponentManager::handleStateQuery - don't wanna be server");
		    }
		}
		break;
	    }
	}
    }

    void checkClientState() {
	if (lj.getClientStateManager().getState() == ClientStateManager.CLIENT_COMPLETE) return;

	Subcomponent sub;
	int state;

	for (Enumeration e = subcomponents.elements(); e.hasMoreElements(); ) {
	    state = ((Subcomponent)e.nextElement()).getState();

	    if (state == Subcomponent.SUBCOMP_DISCOVERED) return; // clientState still incomplete
	    if (state == Subcomponent.REQUESTED) return;
	    if (state == Subcomponent.POLICY_SET) return;
	}

	lj.getClientStateManager().setState(ClientStateManager.CLIENT_COMPLETE);
	lj.getTimerManager().setServerTimeoutTimer();

	if (debug) {
	    System.out.println("SubcomponentManager::checkClientState - client complete");
	}
    }

    void addNetCapacitySubcomponent(long id) {
	netCapSubcomponents.addElement(new Long(id));

	if (netCapSubcomponents.size() == 1) // queue was empty
	    lj.getTimerManager().setNetQueryTimer(true); // first capacity timer
	//else timer running
    }

    void deleteNetCapacitySubcomponent(long id) {
	int i = netCapSubcomponents.indexOf(new Long(id));

	if (i == -1) return;  // Subcomponent not found

	netCapSubcomponents.remove(i);

	if (netCapSubcomponents.size() == 0) // queue is empty now
	    lj.getTimerManager().deleteNetQueryTimer();
    }

    void requestNetCapacitySubcomponent() {
	if (netCapSubcomponents.isEmpty()) return;

	//if (lj.getLJChannelManager.checkNetCapacity()) {
	long id = ((Long)netCapSubcomponents.firstElement()).longValue();
	lj.getLJChannelManager().requestSubcomponentState(id);
	lj.getTimerManager().setStateQueryTimer(id);
	netCapSubcomponents.removeElementAt(0);
	// } else do nothing

    }

    boolean moreNetCapacitySubcomponents() {
	if (netCapSubcomponents.size() >= 1)
	    return true;
	return false;
    }
}






