/* The Java RTP/I latejoin service, Version 0.1 alpha.
 *  
 * This library provides the functionality of a generic RTP/I latejoin service.
 *
 * Copyright (C) 2000 Juergen Vogel, Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Juergen Vogel
 * Martin Mauve
 *
 * e-mail:
 * {vogel,mauve}@informatik.uni-mannheim.de
 * 
 * paper mail:
 *
 * Juergen Vogel
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.services.latejoin;

import java.util.Hashtable;

class TimerManager implements ExecuteTimer {
    private static boolean debug = false;

    // timer computing const's -> SRM
    private static double C1 = 2.0;
    private static double C2 = 4.0;
    private static double D1 = 1.0;
    private static double D2 = 2.0;

    private long maxStateTransmissionTime;

    private Hashtable subcompStateQuery;
    private Hashtable subcompStateReply;
    private boolean serverTimeoutTimerRunning;

    private LateJoin lj;

    TimerManager(LateJoin l, long maxSTT) {
	subcompStateQuery = new Hashtable();
	subcompStateReply = new Hashtable();
	serverTimeoutTimerRunning = false;
	lj = l;
	maxStateTransmissionTime = maxSTT;
    }


    void quit() {
	TimerQueue.deleteTimers(this);
    }

    void setStateQueryTimer(long id) {
	if (subcompStateQuery.get(new Long(id)) == null)
	    subcompStateQuery.put(new Long(id), new StateQueryCounter());

	StateTimerObject timer = new StateTimerObject(computeStateQueryRunTime(id),
						      TimerObject.STATE_QUERY, id, this);
	if (debug) {
	    System.out.println(System.currentTimeMillis()
			       + " TimerManager::setStateQueryTimer - " +
			       "new timer for subcomponent " + id);
	}

	TimerQueue.setNewTimer(timer);
    }


    void deleteStateQueryTimer(long id) {
	subcompStateQuery.remove(new Long(id));
	TimerQueue.deleteTimer(TimerObject.STATE_QUERY, id, this);

	if (debug) {
	    System.out.println("TimerManager::deleteStateQueryTimer - " +
			       "delete timer for subcomponent " + id);
	    TimerQueue.print();
	    System.out.println();
	}
    }


    void resetStateQueryTimer(long id) {
	StateQueryCounter counter = (StateQueryCounter)subcompStateQuery.get(new Long(id));
	if (counter == null) return;          // no timer running

	TimerQueue.deleteTimer(TimerObject.STATE_QUERY, id, this);
	counter.countRuns++;

	if (debug) {
	    System.out.println(System.currentTimeMillis()
			       + " TimerManager::resetStateQueryTimer - tries " + counter.countRuns);
	}

	setStateQueryTimer(id);
    }


    boolean stateQueryTimerRunning(long id) {
	if (subcompStateQuery.get(new Long(id)) != null) return true;
	return false;
    }


    void setStateQueryAcknowledged(long id) {
	StateQueryCounter counter = (StateQueryCounter)subcompStateQuery.get(new Long(id));
	if (counter == null) return;          // no timer running

	counter.acknowledged = true;
    }


    void setStateReplyTimer(long id, int ssrc) {
	StateTimerObject timer = new StateTimerObject(computeStateReplyRunTime(ssrc),
						      TimerObject.STATE_REPLY, id, this);

	if (debug) {
	    System.out.println(System.currentTimeMillis()
			       + " TimerManager::setStateReplyTimer");
	}

	TimerQueue.setNewTimer(timer);

	subcompStateReply.put(new Long(id), new Long(id));
    }


    void deleteStateReplyTimer(long id) {
	subcompStateReply.remove(new Long(id));
	TimerQueue.deleteTimer(TimerObject.STATE_REPLY, id, this);

	if (debug) {
	    System.out.println("TimerManager::deleteStateReplyTimer - " +
			       "delete timer for subcomponent " + id);
	    TimerQueue.print();
	    System.out.println();
	}
    }


    boolean stateReplyTimerRunning(long id) {
	if (subcompStateReply.get(new Long(id)) != null) return true;
	return false;
    }


    void setNetQueryTimer(boolean first) {
	long runningTime;
	if (first)
	    runningTime = (Math.round(Math.random() * 1000));
	else
	    runningTime = 10000;  // 10 secs

	TimerObject timer = new TimerObject(runningTime, TimerObject.NET_QUERY, this);
	TimerQueue.setNewTimer(timer);

	if (debug) {
	    System.out.println(System.currentTimeMillis()
			       + " TimerManager::setNetQueryTimer");
	}
    }


    void deleteNetQueryTimer() {
	TimerQueue.deleteTimer(TimerObject.NET_QUERY,this);

	if (debug) {
	    System.out.println("TimerManager::deleteNetQueryTimer - delete timer");
	    TimerQueue.print();
	    System.out.println();
	}
    }


    void setServerTimeoutTimer() {
	if (serverTimeoutTimerRunning) return;

	MembershipManager mm = lj.getMembershipManager();
	TimerObject timer = new TimerObject(mm.computeTimeout(),TimerObject.SERVER_TIMEOUT, this);

	TimerQueue.setNewTimer(timer);

	serverTimeoutTimerRunning = true;

	if (debug) {
	    System.out.println("TimerManager::setServerTimeoutTimer");
	    TimerQueue.print();
	    System.out.println();
	}
    }


    void deleteServerTimeoutTimer() {
	if (!serverTimeoutTimerRunning) return;

	TimerQueue.deleteTimer(TimerObject.SERVER_TIMEOUT,this);

	serverTimeoutTimerRunning = false;

	if (debug) {
	    System.out.println("TimerManager::deleteServerTimeoutTimer");
	    TimerQueue.print();
	    System.out.println();
	}
    }


    void setClientTimeoutTimer() {
	ClientStateManager cm = lj.getClientStateManager();
	TimerObject timer = new TimerObject(cm.computeTimeout(),TimerObject.CLIENT_TIMEOUT, this);

	TimerQueue.setNewTimer(timer);

	if (debug) {
	    System.out.println("TimerManager::setClientTimeoutTimer");
	    TimerQueue.print();
	    System.out.println();
	}
    }

    void resetClientRequestCounter(long subID) {
	StateQueryCounter counter = (StateQueryCounter)subcompStateQuery.remove(new Long(subID));
	if (counter==null) // subcomponent has been received or reported as failed!
	    return;
	subcompStateQuery.put(new Long(subID), new StateQueryCounter()); // reset counter
    }


    public void executeTimer(TimerObject timer) {
	switch(timer.type) {
	case TimerObject.SERVER_TIMEOUT:
	    {
		lj.messages().put(new LateJoinMessage(LateJoinMessage.TIMER_SERVER_TIMEOUT,
						      timer));
		break;
	    }
	case TimerObject.STATE_QUERY:
	    {
		lj.messages().put(new LateJoinMessage(LateJoinMessage.TIMER_STATE_QUERY,
						      timer));
		break;
	    }
	case TimerObject.STATE_REPLY:
	    {
		lj.messages().put(new LateJoinMessage(LateJoinMessage.TIMER_STATE_REPLY,
						      timer));
		break;
	    }
	case TimerObject.NET_QUERY:
	    {
		lj.messages().put(new LateJoinMessage(LateJoinMessage.TIMER_NET_QUERY,
						      timer));
		break;
	    }
	case TimerObject.CLIENT_TIMEOUT:
	    {
		lj.messages().put(new LateJoinMessage(LateJoinMessage.TIMER_CLIENT_TIMEOUT,
						      timer));
		break;
	    }
	}
    }


    void handleStateQueryTimer(StateTimerObject timer) {
	StateQueryCounter counter = (StateQueryCounter)subcompStateQuery.get(new Long(timer.id));
	SubcomponentManager sm = lj.getSubcomponentManager();

	if (counter == null) // State packet was faster in message queue
	    return;

	if (debug) {
	    System.out.println("TimerManager::handleStateQueryTimer - " +
			       "timer for Subcomponent with ID " + timer.id +
			       " ran " + counter.countRuns + " times out");
	    TimerQueue.print();
	    System.out.println();
	}

	if (counter.countRuns <= lj.initialRequests) {  // request on lj mc group
	    lj.getLJChannelManager().requestSubcomponentState(timer.id);
	    counter.countRuns++;
	    setStateQueryTimer(timer.id);
	    return;
	}
	
	if (counter.countRuns < lj.initialRequests+lj.joinRequests) {
	    // request on data generic service control channel
	    int tries = counter.countRuns - lj.initialRequests; // tries on control channel
	    lj.getControlChannelManager().requestSubcomponentState(timer.id, tries);

	    if (counter.countRuns == lj.initialRequests + 1)
		counter.acknowledged = false; // reset on first try
	    counter.countRuns++;
	    setStateQueryTimer(timer.id);
	    return;
	}

	// query failed
	subcompStateQuery.remove(new Long(timer.id));
	sm.stateQueryFailed(timer.id);
    }


    void handleStateReplyTimer(StateTimerObject timer) {
	lj.getLateJoinRecipient().lateJoinGetState(timer.id);
	lj.getSubcomponentManager().getSubcomponent(timer.id).setLastTimeSeenNow();
	subcompStateReply.remove(new Long(timer.id));

	if (debug) {
	    System.out.println("TimerManager::handleStateReplyTimer - " +
			       "timer for Subcomponent with ID " + timer.id + " ran out");
	    TimerQueue.print();
	    System.out.println();
	}
    }


    void handleNetQueryTimer(TimerObject timer) {
	if (debug)
	    System.out.println("TimerManager::handleNetQueryTimer - timer ran out");

	SubcomponentManager sm = lj.getSubcomponentManager();
	sm.requestNetCapacitySubcomponent();

	if (sm.moreNetCapacitySubcomponents())
	    setNetQueryTimer(false);
    }


    void handleServerTimeoutTimer(TimerObject timer) {
	lj.getMembershipManager().reset();
	lj.getLJChannelManager().leaveGroup();

	serverTimeoutTimerRunning = false;

 	if (debug) {
	    System.out.println("TimerManager::handleServerTimeoutTimer - timer ran out");
	    TimerQueue.print();
	    System.out.println();
	}
    }


    void handleClientTimeoutTimer(TimerObject timer) {
	lj.getClientStateManager().setLearning(false);

 	if (debug) {
	    System.out.println("TimerManager::handleClientTimeoutTimer - timer ran out");
	    TimerQueue.print();
	    System.out.println();
	}
    }


    private long computeStateQueryRunTime(long id) {
	StateQueryCounter counter = (StateQueryCounter)subcompStateQuery.get(new Long(id));
	long runningTime;

	if (counter.countRuns <= lj.initialRequests) {
	    // StateQuery triggers anycast to select a server of the lj group
	    double rTime;
	    double delay;

	    if (counter.acknowledged) {
		// we have an server now, the StateQueryTimer ensures the reliable transmission
		// of the state -> the runningTime has to allow for the longer transmission time
		// of the state packet
		delay = lj.getDelayManager().getAverageDelay() + maxStateTransmissionTime;

		if (debug)
		    System.out.println("computeStateQueryRunTime - ljGroup && ack");
	    } else {
		// we have no server yet, the StateQueryTimer ensures the reliable transmsission
		// of the StateQuery -> no additional transmission time needed
		delay = lj.getDelayManager().getAverageDelay(); // delay between me and rest of group

		if (debug)
		    System.out.println("computeStateQueryRunTime - ljGroup && !ack");
	    }

	    rTime =  delay*(C1 + Math.random()*C2); // see SRM
	    //rTime *= Math.pow(2.0, counter.countRuns - 1.0); we don't need this

	    runningTime = Math.round(rTime);

	} else {
	    // StateQuery triggers anycast to select a server for joining the lj group no randomization since it should already be random!

	    if (counter.acknowledged) {
		// we have an server now, the StateQueryTimer ensures the reliable transmission
		// of the state -> the runningTime has to allow for the longer transmission time
		// of the state packet
		runningTime = 2 * lj.getDelayManager().getMaxDelay() + maxStateTransmissionTime;

		if (debug)
		    System.out.println("computeStateQueryRunTime - controlChannel && ack");
	    } else {
		// we have no server yet, the StateQueryTimer ensures the reliable transmsission
		// of the StateQuery -> no additional transmission time needed
		runningTime = 2 * lj.getDelayManager().getMaxDelay();

		if (debug)
		    System.out.println("computeStateQueryRunTime - controlChannel && !ack");
	    }

	}

	if (runningTime < 30) return Math.round(30+Math.random()*30d);           // no timer less than 30ms
	else return runningTime;
    }


    private long computeStateReplyRunTime(int ssrc) {
	double runningTime;
	double delay = lj.getDelayManager().getDelay(ssrc); // delay between me and query source
	
	runningTime = delay*(D1 + Math.random()*D2); // see SRM

	if (runningTime < 10) return Math.round(10+Math.random()*20d);           // no timer less than 10ms
	return Math.round(runningTime);
    }
}


