/* The Java RTP/I latejoin service, Version 0.1 alpha.
 *  
 * This library provides the functionality of a generic RTP/I latejoin service.
 *
 * Copyright (C) 2000 Juergen Vogel, Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Juergen Vogel
 * Martin Mauve
 *
 * e-mail:
 * {vogel,mauve}@informatik.uni-mannheim.de
 * 
 * paper mail:
 *
 * Juergen Vogel
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.services.latejoin;

import java.util.*;

class TimerQueue {
    private static final boolean debug = false;

    private static Vector queue = new Vector();
    private static TimerThread runningTimer;
    static boolean threadRunning = false;

    public static synchronized void setNewTimer(TimerObject timer) {

	// search position in queue
	int pos = 0;
	while ((pos < queue.size())
	       && timer.greaterThan((TimerObject)queue.elementAt(pos))) pos++;
	queue.insertElementAt(timer, pos);

	if (debug) {
	    System.out.print("TimerQueue::setNewTimer - insert timer ");
	    timer.print();
	    System.out.println(" at Position " + pos);
	}

	// 1. no timer running - inserted timer is the only one in the queue
	if (!threadRunning) {
	    if (debug) {
		System.out.println(System.currentTimeMillis() +
				   " TimerQueue::setNewTimer - no thread running");
	    }
	    setTimerThread(); // start inserted timer
	    return;
	}

	// 2. timer running
	// 2.1 inserted timer t1 has an earlier ending time than running timer t0
	if (pos == 0 && runningTimer.timer.greaterThan(timer)) {
	    runningTimer.interrupt(); // delete running thread
	    threadRunning = false;
	    queue.insertElementAt(runningTimer.timer,1); // insert t0 after t1
	    if (debug) {
		System.out.println(System.currentTimeMillis() +
				   " TimerQueue::setNewTimer - reset running thread ");
	    }
	    setTimerThread(); // start t1
	    return;
	}

	// 2.2 inserted timer has a later ending time than running timer
	// nothing to do
	if (debug) {
	    System.out.println(System.currentTimeMillis() +
			       " TimerQueue::setNewTimer - timer is later than thread running");
	}
   }

    public static synchronized void print() {
	runningTimer.timer.print();
	System.out.print(" - ");
	for (int i = 0; i < queue.size(); i++)
	    ((TimerObject)queue.elementAt(i)).print();
    }

    public static synchronized void deleteTimer(int type, ExecuteTimer callback) {
	// 1. timer to delete is running timer
	if ((runningTimer.timer.type == type) && (runningTimer.timer.callback == callback)) {
	    if (debug) {
		System.out.println("TimerQueue::deleteTimer - reset TimerThread");
	    }
	    runningTimer.interrupt(); // delete running timer
	    threadRunning = false;
	    setTimerThread(); // set new timer
	    return;
	}

	// 2. timer to delete is in queue
	for (int i = 0; i < queue.size(); i++)
	    if ((((TimerObject)queue.elementAt(i)).type == type) && (((TimerObject)queue.elementAt(i)).callback==callback)) {
		queue.removeElementAt(i);
		return;
	    }

	// 3. timer to delete not found
	return;
    }

    public static synchronized void deleteTimer(int type, long id, ExecuteTimer callback) {
	// 1. timer to delete is running timer
	if ((runningTimer.timer.type == type) && 
	    (runningTimer.timer.callback == callback) &&
	    (((StateTimerObject)runningTimer.timer).id == id)) {
	    if (debug) {
		System.out.println("TimerQueue::deleteTimer - reset TimerThread");
	    }
	    runningTimer.interrupt(); // delete running timer
	    threadRunning = false;
	    setTimerThread(); // set new timer
	    return;
	}

	// 2. timer to delete is in queue
	for (int i = 0; i < queue.size(); i++)
	    if ((((TimerObject)queue.elementAt(i)).type == type) && 
		(((TimerObject)queue.elementAt(i)).callback == callback) && 
		(((StateTimerObject)queue.elementAt(i)).id == id)) {
		queue.removeElementAt(i);
		return;
	    }

	// 3. timer to delete not found
	return;
    }

    public static synchronized void deleteTimers(ExecuteTimer callback) {
	// 1. timer to delete is running timer
	if (runningTimer.timer.callback == callback) {
	    if (debug) {
		System.out.println("TimerQueue::deleteTimers - reset TimerThread");
	    }
	    runningTimer.interrupt(); // delete running timer
	    threadRunning = false;
	}

	// 2. timer to delete is in queue
	for (int i = 0; i < queue.size(); i++)
	    if (((TimerObject)queue.elementAt(i)).callback == callback) { 
		queue.removeElementAt(i);
		return;
	    }
	
	if (!threadRunning) {
	    setTimerThread();
	}
	return;
    }	

    public static synchronized void setTimerThread() {
	if (threadRunning || (queue.size() == 0)) return;

	threadRunning = true;

	TimerObject timer = (TimerObject)queue.firstElement();
	long sleepingTime = timer.computeSleepingTime();

	// execute all timer with a remaining running time of less than 10 milliseconds
	while (sleepingTime < 10) {
	    if (debug) {
		System.out.println(System.currentTimeMillis()
				   + " TimerQueue::setTimerThread - execute ");
	    }
	    if (timer.callback!=null) {
		timer.callback.executeTimer(timer);
	    }
	    queue.removeElementAt(0);
	    if (queue.size() == 0) return; //all timer done
	    timer = (TimerObject)queue.firstElement();
	    sleepingTime = timer.computeSleepingTime();
	}


	runningTimer = new TimerThread(timer, sleepingTime);
	queue.removeElementAt(0);
	if (debug) {
	    System.out.println(System.currentTimeMillis() +
			       " TimerQueue::setTimerThread - new Thread");
	}
	runningTimer.start();
    }
}

