/* The Java RTP/I library, Version 0.1 alpha.
 *  
 * This library provides the functionality of RTP/I as it is specified in
 * Internet Draft draft-mauve-rtpi-00.txt. 
 *
 * Copyright (C) 2000 Martin Mauve
 * University of Mannheim / Germany
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Martin Mauve
 *
 * e-mail:
 * mauve@informatik.uni-mannheim.de
 * 
 * paper mail:
 * Martin Mauve
 * University of Mannheim
 * Lehrstuhl Praktische Informatik IV
 * L15, 16
 * 68131 Mannheim
 * Germany
 */

package rtpi.transport;

import java.io.IOException;

/**
 * This is the interface which needs to be implemented by a transport
 * layer implementation. By implementing Transport, the transport
 * layer can be used with the rtp and rtpi libraries.
 *
 * @author Martin Mauve
 */

public interface Transport {

    /**
     * This terminates the transport service
     */

    void quit();

    /**
     * Register a recipient for the packets received by this tranport
     * instance. Transport packets are handed only to the last 
     * TransportRecipient instance for which this method is called.
     *
     * @param transRec The transport recipient which should receive
     *                 incoming transport packets.
     */

    void registerTransportRecipient(TransportRecipient transRec);


    /**
     * Join the transport session. For multicast transport implementations
     * this is call joins the multicast group. Unicast implementations
     * should make sure that no packets are handed to the TransportRecipien
     * before this method is called.
     */

    void joinGroup() throws IOException;

    /**
     * Leave the transport session. For multicast transport implementations
     * this call leaves the multicast group. Unicast implementations
     * should make sure that no packets are handed to the TransportRecipient
     * after this method is called.
     */

    void leaveGroup();


    /**
     * Sends a TransportPacket.
     *
     * @param packet The packet to be sent.
     */

    void sendTransportPacket(TransportPacket packet);


    /**
     * Get the header size for this transport (network+transport header!)
     *
     * @return The header size.
     */

     int getHeaderSize();

    /**
     * Returns the maximum payload size which can be transported
     * without fragmentation by the transport implementation.
     *
     * @return the maximum payload size
     */
 
    int getTransportPayloadSize();


    /**
     * This requests that a NotificationRecipient is informed when the 
     * network becomes idle.
     *
     * @param sendFree The boundary of the send data rate (bit/s) below which the network
     *                   is considered idle.
     * @param receivFree The boundary of of the received data rate (bit/s) below which the
     *                   network is considered idle. This includes the data send. Therefore:
     *                   receivedFree=sendFree+x  
     * @param  recipient The notificationRecipient that is interested in receiving a 
     *                   notification when the network becomes idle.
     */ 

    void queryNetworkResources(int senderFree, int receiverFree, NotificationRecipient recipient);

}


